/*
                * logWebObjects() -- var arg logging utility for debuggin
 */ 
#include <stdio.h>
#include <stdlib.h>
#ifdef WIN32
#include <sys/param.h>
#include <direct.h>
#include <winsock.h>
#include <errno.h>
#include <sys/types.h>
#include <signal.h>
#include <setjmp.h>
#else
#include <unistd.h>
#include <netdb.h>
#endif

#include <stdarg.h>
#include <sys/stat.h>
#include <ctype.h>

#ifdef NeXT_PDO
#include <string.h>
#else
#include <strings.h>
#endif

#ifdef NeXT
#include <libc.h>
#endif

#if (defined(NeXT_PDO) && defined(__svr4__))
#include <sys/param.h>
#endif

#if  (NeXT || (NeXT_PDO && __svr4__)) && !defined(WIN32)
extern int gethostname(char *hn, int size); /* can't find this anywhere */
#endif

#include "WOAdaptorUtil.h"
#include "WODict.h"
#include "WOConfigFile.h"

#define QUOTE '"'
#define MAX_ERROR_LENGTH 1024
#define ERROR_HEAD "Content-Type: text/html\n\n<html><body><strong>"
#define ERROR_TAIL "</strong></body></html>\n"

char *WO_CGI_Headers[32] = {
                CGI_REQUEST_METHOD,
                CGI_SERVER_SOFTWARE,
                CGI_SERVER_NAME,
                CGI_GATEWAY_INTERFACE,
                CGI_SERVER_PORT,
                CGI_QUERY_STRING,
                CGI_REMOTE_HOST,
                CGI_REMOTE_ADDR,
                CGI_REMOTE_USER,
                CGI_REMOTE_IDENT,
                CGI_AUTH_TYPE,
                CGI_AUTH_PASS,
      		CGI_ANNOTATION_SERVER,
                CGI_CONTENT_TYPE,
                CGI_CONTENT_LENGTH,
                CGI_HTTP_USER_AGENT,
                CGI_HTTP_DATE,
                CGI_HTTP_MIME_VERSION,
                CGI_HTTP_PRAGMA,
                CGI_HTTP_FROM,
                CGI_HTTP_IF_MODIFIED_SINCE,
                CGI_HTTP_REFERER,
                CGI_HTTP_ACCEPT,
                CGI_HTTP_ACCEPT_ENCODING,
                CGI_HTTP_ACCEPT_LANGUAGE,
                CGI_HTTP_ALLOW,
                CGI_HTTP_CONTENT_ENCODING,
                CGI_HTTP_EXPIRES,
                CGI_HTTP_LAST_MODIFIED,
                CGI_HTTP_SET_COOKIE,
                CGI_HTTP_AUTHORIZATION,
                NULL};

char *WO_Req_Headers[32] = {
                "x-webobjects-request-method",
                "x-webobjects-server-software",
                "x-webobjects-server-name",
                "x-webobjects-gateway-interface",
                "x-webobjects-server-port",
                "x-webobjects-query-string",
                REMOTE_HOST,
                REMOTE_ADDR,
                REMOTE_USER,
                "x-webobjects-remote-ident",
                "x-webobjects-auth-type",
                "x-webobjects-auth-pass",
    		"x-webobjects-annotation-server",
                CONTENT_TYPE,
                CONTENT_LENGTH,
                USER_AGENT,
                DATE,
                MIME_VERSION,
                PRAGMA,
                FROM,
                IF_MODIFIED_SINCE,
                REFERER,
                ACCEPT,
                ACCEPT_ENCODING,
                ACCEPT_LANGUAGE,
                ALLOW,
                CONTENT_ENCODING,
                EXPIRES,
                LAST_MODIFIED,
                SET_COOKIE,
                AUTHORIZATION,
                NULL};

static int attemptedToLogOnce = 0;
static int shouldLog = 0;

void logWebObjects(char *name, char *fmt, ...)
{
    if (!attemptedToLogOnce) {
	struct stat statbuf;
    	char *fname;
#ifndef WIN32
	fname = "/tmp/logWebObjects";
#else
        char tempdir[WO_MAX_PATH+1];
        WOReadKeyFromConfiguration ("TEMP", tempdir, WO_MAX_PATH);
	fname = (char*)malloc (strlen (tempdir) + strlen ("/logWebObjects") + 1);
	sprintf (fname, "%s/logWebObjects", tempdir);
#endif
    	attemptedToLogOnce = 1;
	if (!stat(fname,&statbuf) && (statbuf.st_uid == 0)) {
	    shouldLog = 1;
	}
#ifdef WIN32
	free(fname);
#endif

    }
		
    if (shouldLog) {
	char *msgLog;
	FILE *log;
	va_list ap;
#ifndef WIN32
	msgLog = "/tmp/WebObjects.log";
#else
        char tempdir[WO_MAX_PATH+1];
        WOReadKeyFromConfiguration ("TEMP", tempdir, WO_MAX_PATH);
	msgLog = (char*)malloc (strlen (tempdir) + strlen ("/WebObjects.log") + 1);
        sprintf (msgLog, "%s/WebObjects.log", tempdir);
#endif
	va_start(ap, fmt);    
	log = fopen(msgLog,"a");
	if (log) {
	    fprintf(log,"%s: ",name);
	    vfprintf(log,fmt,ap);
	    fprintf(log,"\n");
	    fflush(log);
	    fclose(log);
	}
	va_end(ap);
#ifdef WIN32
	free(msgLog);
#endif
    }
    
}

void WOBreakUpCookies(const char *cookies, dict *dictionary) {
    char *cookieString;
    int len;

    if (cookies && strlen(cookies)) {
	char *start, c, *p;

	len = strlen(cookies);
	cookieString = malloc(len+1);
	strcpy(cookieString,cookies);
	
	p = cookieString;
	start = cookieString;

	while ( (c = *p) ) {
	    if (c == ';') {
		*p = '\0';
		++p;
		if (start && *start) {
		    dict_nvinsert("Cookie", start, dictionary);
		}
		while (*p && ((*p == ' ') || (*p == '\t'))) ++p;
		start = p;
	    } else {
		++p;
	    }
	}
	if (start && *start) {
	    dict_nvinsert("Cookie", start, dictionary);
	}
	free(cookieString);
    }
}


/*
 * ISMap() check to see if the user input is from
 * an ISMAP image map -- i.e. check to see if it's of the format 
 * "number,number"
 * check if there is only one comma, if the comma is not at an extremity
 * and that all the other characters are digits.
 */
int ISMAP(char *buf)
{
    char c,  *p=buf, *comma;
    int commaCount = 0;
    
    comma = strchr(buf,',');
    if (!comma || (comma[1]=='\0') || (comma == buf)) { 
	return (0);		/* There is no comma, or it is at the beginning or at the end of the string */
    	}

    while ((c = *p++)) {
	if (c==',') commaCount++;
	if  ((commaCount > 1) ||  (!isdigit(c) && (c != ','))) {
		return (0); /*There is only one comma, all others are digits */
		}
	}  
	
    return (1);
}

char *lowerCaseString(char *input) {
    char *p, *q, c;
    
    p = q = input;
    while ( (c = *p++) ) {
	*q++ = ((c >= 'A') && (c <='Z')) ? tolower(c) : c;
    }
    return input;
}

char *WOStripNL(char *input) {
    char *p, c;
    int len;
    if (input && (len = strlen(input))) {
	p = &input[len-1];
	while (len-- && (c = *p) && ((c =='\n') || (c== '\r'))) {
	    *p-- = '\0';
	}
    }
    return input;
}


#ifdef WIN32
char *getLineFromSocket(char *buf, int size, int sd) {
	char c[1];
	int count = 0;
	do
	{
            recv (sd, c, sizeof (char), 0);
            buf[count++] = c[0];
	} while ((c[0] != '\n') && (count < size));
	--count;
	while (buf[count] == '\r' || buf[count] == '\n')  
		buf[count--] = '\0';

	if (buf[0])
		return buf;
	else
		return NULL;
}
#endif

char *getLineFromFilePointer(char *buf, int size, FILE *fp) {
    char *p, *r = fgets (buf,size,fp);

    if (r != NULL) {
        p = buf + strlen(buf) - 1;
        while (p >= buf && (*p == '\r' || *p == '\n'))
            *p-- = '\0';
        return buf;
    } else {
        return NULL;
    }
}

char *getLine(char *buf, int size, FILE *fp) {
#ifdef WIN32
    return getLineFromSocket(buf, size, (int)fp);
#else
    return getLineFromFilePointer(buf, size, fp);
#endif
}

/*****************************************************************************
*                          Registry data  functions                          *
*****************************************************************************/
/*
 * This function is used on Windows to get the information out of the
 * registration database.
 */

#ifdef WIN32
const char *WOReadKeyFromConfiguration (char *keyName, char *buffer, int bufferSize)
{
    HKEY newHandle;
    LONG result;
    REGSAM security = KEY_READ;
        DWORD subKeys;
        DWORD maxSubKeyLength;
        DWORD maxClassLength;
        DWORD values;
        DWORD maxValueNameLength;
        DWORD maxValueDataLength;
        DWORD securityDescriptor;
        FILETIME fileTime;

    result = RegOpenKeyEx( HKEY_LOCAL_MACHINE, "SOFTWARE\\NeXT\\WebObjects\\Configuration", 0, security, &newHandle );
        if (result == ERROR_SUCCESS)
        {
            DWORD classSize = WO_MAX_PATH+1;
            char  className[classSize-1];
            DWORD type;

            RegQueryInfoKey( newHandle, className, &classSize, NULL,
                    &subKeys, &maxSubKeyLength, &maxClassLength, &values,
                    &maxValueNameLength, &maxValueDataLength,
                    &securityDescriptor, &fileTime );
            memset((char *)buffer,0,bufferSize);
            if( RegQueryValueEx( newHandle, keyName, 0, &type, buffer, (LPDWORD)&bufferSize ) == ERROR_SUCCESS ) {
                return buffer;
            }
        }
        buffer[0] = '\0';
        return buffer;
}
#endif

/******************************************************************************/

/*****************************************************************************
*                          Current Working Directory functions               *
*****************************************************************************/
/*
 * This function is used on all platforms to get the current working directory
 * It returns a malloc'ed char*, that you are responsible for freeing.
 */

char *WOGetCurrentWorkingDirectory() {
    char *path = NULL;

    path = malloc((WO_MAX_PATH+1)*sizeof(char));
#if (defined(NeXT_PDO) && (defined(__svr4__) || defined(hpux) || defined(HPUX) || defined(hppa)))
    getcwd(path,WO_MAX_PATH+1);	// HPUX
#else
#ifdef WIN32
    _getcwd(path, WO_MAX_PATH+1);	// Windows
#else
    getwd(path);		// NeXTSTEP
#endif
#endif 

return path;
}

/******************************************************************************/

/*
 * This function is used to open (read-only) the public configuration file.
 * You are responsible for closing the file pointer if opened.
 */

FILE *WOOpenConfigFile(const char *configFile) {

/* As of WOF3.1, we look at a constant location for the Configuration File. This will simplify deployment.
* the constant location for the conf file is <NEXT_ROOT>/NextLibrary/WOAdaptors/Configuration/WebObjects.conf
*
*/
    FILE *fp = NULL;

    if (!(fp = fopen(configFile, "r")) ) {
        logWebObjects(LOGCGI_WARNING,"-- OpenConfigFile: \"%s\" occured while opening the configuration file %s .\n",strerror(errno),configFile);
        logWebObjects(LOGCGI_WARNING,"-- OpenConfigFile: Trying to open obsolete configuration file.");
        if (!(fp = fopen(CONFIG_FILE,"r"))) 
        {
/* Get the current working directory for any platform (for logging purpose).
*  We are responsible for free'ing it.
*/
            char *path = WOGetCurrentWorkingDirectory();
            logWebObjects(LOGCGI_WARNING,"-- OpenConfigFile: \"%s\" occured while opening the configuration file %s%s%s\n",strerror(errno),path ? path : "", path ? "/" : "",CONFIG_FILE);
            if (path) free(path);
         }
	
    }
    return fp;
}
