#include <stdio.h>
#include <errno.h>
#include <string.h>
#include <stdlib.h>
#include "WODict.h"

/*
 * Extension definitions
 */

#define WOADAPTOR_APP_EXTENSION ".woa"
#define WOADAPTOR_APP_EXTENSION_UPPERCASE ".WOA"
#define WOADAPTOR_COMPONENT_EXTENSION ".wo"
#define WOADAPTOR_COMPONENT_EXTENSION_UPPERCASE ".WO"

/*
 * Max Path Length definition
 */

#ifdef NeXT
#include <libc.h>
#define WO_MAX_PATH MAXPATHLEN
#endif

#if (defined(NeXT_PDO) && defined(__svr4__))
#define WO_MAX_PATH MAXPATHLEN
#endif

#if (defined(NeXT_PDO) && (defined(hpux) || defined(HPUX) || defined(hppa)))
#include <limits.h>
#define WO_MAX_PATH PATH_MAX
#endif

#if  defined(WIN32)
#define WO_MAX_PATH MAX_PATH
#endif


/*
 * error html files definition
 */

#define ERROR_RESPONSE_LOCKED_AUTOSTART_FILE	"/WebObjects/info/autostart_error1.html"
#define ERROR_RESPONSE_WAITED_TOO_LONG		"/WebObjects/info/autostart_error2.html"
#define ERROR_RESPONSE_CANNOT_LOCK_CONFIG	"/WebObjects/info/autostart_error3.html"
#define ERROR_RESPONSE_CANNOT_OPEN_CONFIG	"/WebObjects/info/autostart_error4.html"
#define ERROR_RESPONSE_PORT_NOT_IN_CONFIG	"/WebObjects/info/autostart_error5.html"
#define ERROR_RESPONSE_NO_RECORDING_FILE	"/WebObjects/info/recording_error1.html"

/* 
 * http method definitions
 */
#define HTTP_NO_METHOD		-2
#define HTTP_UNKNOWN_METHOD	-1
#define HTTP_GET_METHOD		0
#define HTTP_POST_METHOD	1
#define HTTP_HEAD_METHOD	2
#define HTTP_PUT_METHOD		3

#define GET_METHOD		"GET"
#define POST_METHOD		"POST"
#define HEAD_METHOD		"HEAD"
#define PUT_METHOD		"PUT"

#define LOGCGI_STAT	"STAT"
#define LOGCGI_WARNING	"WARN"
#define LOGCGI_INFO	"INFO"

/* this defines how many bytes we should read at a time when
 * using sockets to read the HTTP data headers
 */
#define HTTP_READ_LENGTH 1
#define HTTP_DATA
#define HUGE_STRING_LENGTH 8192
#define HUGE_BUFFER	8192

/* Our own custom header for the Document Root */

#define DOCUMENT_ROOT "x-webobjects-documentroot"

/* Our own custom headers for client info */

#define REMOTE_ADDR "x-webobjects-remote-addr"
#define REMOTE_HOST "x-webobjects-remote-host"
#define REMOTE_USER "x-webobjects-remote-user"

/*
 * http 1.0 request protocol headers
 */

/* General Headers */
#define DATE "date"
#define MIME_VERSION "mime-version"
#define PRAGMA "pragma"

/* Request Headers */
#define AUTHORIZATION "authorization"
#define FROM "from"
#define IF_MODIFIED_SINCE "if-modified-since"
#define REFERER "referer"
#define USER_AGENT "user-agent"
#define ACCEPT "accept"
#define ACCEPT_ENCODING "accept-encoding"
#define ACCEPT_LANGUAGE "accept-language"

/* Response Headers */
#define LOCATION "location"
#define SERVER "server"
#define WWW_AUTHENTICATE "www-authenticate"

/* Entity Headers */
#define ALLOW "allow"
#define CONTENT_ENCODING "content-encoding"
#define CONTENT_LENGTH "content-length"
#define CONTENT_TYPE "content-type"
#define DEFAULT_CONTENT_TYPE "application/x-www-form-urlencoded"
#define EXPIRES "expires"
#define LAST_MODIFIED "last-modified"

/* Netscape cookies header */
#define SET_COOKIE "cookie"

/* Status header */
#define STATUS "Status"


/* 
 * CGI Environment vars for CGI 1.1 - Server Specific
 */

#define CGI_SERVER_SOFTWARE	"SERVER_SOFTWARE"
#define CGI_SERVER_NAME		"SERVER_NAME"
#define CGI_GATEWAY_INTERFACE	"GATEWAY_INTERFACE"
#define CGI_SERVER_PROTOCOL	"SERVER_PROTOCOL"
#define CGI_SERVER_PORT		"SERVER_PORT"
#define CGI_REQUEST_METHOD	"REQUEST_METHOD"
#define CGI_PATH_INFO		"PATH_INFO"
#define CGI_PATH_TRANSLATED	"PATH_TRANSLATED"
#define CGI_SCRIPT_NAME		"SCRIPT_NAME"
#define CGI_QUERY_STRING	"QUERY_STRING"
#define CGI_REMOTE_HOST		"REMOTE_HOST"
#define CGI_REMOTE_ADDR		"REMOTE_ADDR"
#define CGI_REMOTE_USER		"REMOTE_USER"
#define CGI_REMOTE_IDENT	"REMOTE_IDENT"
#define CGI_AUTH_TYPE		"AUTH_TYPE"
#define CGI_AUTH_PASS		"AUTH_PASS"
#define CGI_CONTENT_TYPE	"CONTENT_TYPE"
#define CGI_CONTENT_LENGTH	"CONTENT_LENGTH"
#define CGI_ANNOTATION_SERVER	"ANNOTATION_SERVER"

/* 
 * CGI Environment vars for CGI 1.1 - HTTP Specific
 * Authorization, Content-Type & Content-Length are excluded
 * as they are already processed.
 */

#define CGI_HTTP_USER_AGENT	"HTTP_USER_AGENT"
#define CGI_HTTP_DATE "HTTP_DATE"
#define CGI_HTTP_MIME_VERSION "HTTP_MIME_VERSION"
#define CGI_HTTP_PRAGMA "HTTP_PRAGMA"
#define CGI_HTTP_FROM "HTTP_FROM"
#define CGI_HTTP_IF_MODIFIED_SINCE "HTTP_IF_MODIFIED_SINCE"
#define CGI_HTTP_REFERER "HTTP_REFERER"
#define CGI_HTTP_ACCEPT "HTTP_ACCEPT"
#define CGI_HTTP_ACCEPT_ENCODING "HTTP_ACCEPT_ENCODING"
#define CGI_HTTP_ACCEPT_LANGUAGE "HTTP_ACCEPT_LANGUAGE"
#define CGI_HTTP_ALLOW "HTTP_ALLOW"
#define CGI_HTTP_CONTENT_ENCODING "CONTENT_ENCODING"
#define CGI_HTTP_EXPIRES "HTTP_EXPIRES"
#define CGI_HTTP_LAST_MODIFIED "HTTP_LAST_MODIFIED"
#define CGI_HTTP_SET_COOKIE "HTTP_COOKIE"
#define CGI_HTTP_AUTHORIZATION "HTTP_AUTHORIZATION"

/* 
 * These are utilitie functions that the adaptor writer can use to speed up
 * the process.
 */

void logWebObjects(char *name, char *fmt, ...);
    /* logging utility -- logs to /tmp/CGIMessenger.log */


int ISMAP(char *buf);
    /* check to see if the user input is from an ISMAP image map */

char *get_application_name(const char *pathInfo, char *buffer, int bufsize);
    /* get_application_name() - get the name of application. The input is pathInfo, which is everything after /cgi-bin/WebObjects/. This name includes the DO session number */

char *get_application_host(const char *pathInfo, char *buffer, int bufsize);
/* get_application_host() - get the host of the application */
/* get the application host from the uri: pathInfo = /cgi-bin/WebObjects/HelloWorld.1@webdrone returns 'webdrone' */

char *get_appname_from_doname(const char *doName, char *appNameBuf, int bufsize);
    /* get_appname_from_doname() - get application name from DO name. This function extracts the appname from the Appname/DO name value (which has the format appName.session#  */

char *lowerCaseString( char *input);

char *WOStripNL(char *input);

char *getLineFromFilePointer(char *buf, int size, FILE *fp);
char *getLine(char *cp, int len, FILE *fp);

void WOSeparateHostNameAndPort(const char *hname, char *hostName, int *port);
void WOSeparateAppNameAndInstance(const char *aName,char *appName, int *ins);

void WOBreakUpCookies(const char *cookieString, dict *dictionary);

#ifdef WIN32
const char* WOReadKeyFromConfiguration (char* keyName, char *buffer, int bufferSize);
#endif /* WIN32 */

char *WOGetCurrentWorkingDirectory();
/* gets the current working directory for all platforms. You are responsible for free'ing the returned char* */

FILE *WOOpenConfigFile(const char *configFile);
/* returns the FILE* for the public Configuration file, at either the new location or the old (pre-WOF3.1) 
 * use fclose() to close the FILE*
*/
