/* WOConfigFile
 */

#include <stdio.h>
#include <stdlib.h>
#ifndef WIN32
#include <unistd.h>
#endif
#include <string.h>
#include <ctype.h>
#include <errno.h>
 
#include "WOConfigFile.h"
#include "WOAdaptorUtil.h"

#if (defined(NeXT_PDO) && (defined(hpux) || defined(HPUX) || defined(hppa)))
#include <limits.h>
#endif

/* returns the port as read from the configuration file.
 * on failure returns 0.
 */

int WOGetPortFromConfigFile(const char *appName, int applicationNumber, const char *hostName, const char *configFile) {

    FILE *fp = NULL;
    char string[MAX_LENGTH_CONF_LINE];
    char app[MAX_LENGTH_CONF_LINE];
    char name[MAX_LENGTH_CONF_LINE];
    int port;
    
    if (hostName) {
	sprintf(string,"%s:%d@%s",appName,applicationNumber,hostName);
    }
    else {
	logWebObjects(LOGCGI_WARNING,"-- WOConfigFile: No host specified. Cannot get port");
	return -1;
    }
	    
    logWebObjects(LOGCGI_INFO,"-- WOConfigFile: Looking for port for %s:%d on host %s",appName, applicationNumber, hostName ? hostName : "localhost");

    if (!(fp = WOOpenConfigFile(configFile))) {
        return -1;
    }
	
    
    while (getLineFromFilePointer(app, MAX_LENGTH_CONF_LINE-1,fp)) {
	if (strstr(app, string)){
	    if (sscanf(app, "%s %d",name,&port)!= 2) {
		logWebObjects(LOGCGI_WARNING,"-- WOConfigFile: Invalid format in configuration file %s : \"%s\" occured while parsing %s\n",CONFIG_FILE,strerror(errno),app);
		fclose(fp);
		return -1;
	    }
	    if (!strcmp(name,string)) {
		fclose(fp);
	    	return port;
	    }	
	}
    }
    if (ferror(fp)) {
	logWebObjects(LOGCGI_WARNING,"-- WOConfigFile: \"%s\" occured while reading the configuration file %s\n",strerror(errno),CONFIG_FILE);
    }
    else {
	logWebObjects(LOGCGI_WARNING,"-- WOConfigFile: Cannot find %s in configuration file %s\n",string,CONFIG_FILE);
    }
    fclose(fp);
    return -1;
        
}
