#include <stdio.h>
#include <stdlib.h>
#ifndef WIN32
#include <unistd.h>
#if defined(NeXT_PDO) && (defined(__alpha__) || defined(hpux))
#include <string.h>
#endif // NeXT_PDO && __alpha__
#endif
#include "WODict.h"

struct dict_entry * _dict_find(char *name, dict *pb)
{
    int i;
    for (i=0;i < pb->hsize;i++) {
	if (pb->ht[i]) {
	    struct dict_entry *entry;
	    entry = pb->ht[i];
	    if (entry->param) {
		if (!strcmp(name,entry->param->name)) {
		    return entry;
		}
	    }
	}
    }
    return NULL;
}

void _insert_entry (struct dict_entry *root, struct dict_entry *entry)
{
    while (root) {
	if (!root->next) {
	    root->next = entry;
	    return;
	}
	root = root->next;
    }
}

void _dict_expand(dict *pb, int nBy) /* expand a dict by nBy */
{
    struct dict_entry **newEntries;
    int i,  newsize = pb->hsize + nBy;
    newEntries = calloc(newsize,sizeof(struct dict_entry));
    for (i=0;i < pb->hsize;i++) {
	newEntries[i] = pb->ht[i];
    }
    free(pb->ht);
    pb->ht = newEntries;
    pb->hsize = newsize;
}


void _dict_insert_entry(struct dict_entry *entry, dict *pb)
{
    int i;
    struct dict_entry *anEntry;
    
    anEntry = _dict_find(entry->param->name, pb);
    if (anEntry) {
	_insert_entry(anEntry,entry);
    } else {
	for (i=0;i < pb->hsize;i++) {
	    if (!pb->ht[i]) {
		pb->ht[i] = entry;
		return;
	    }
	}
	/* we need to expand it */
	_dict_expand(pb, 10);
	/* add new entry to named table */
	pb->ht[i] = entry;
    }
}

void _remove_entry_list(struct dict_entry *entry)
{
    dict_param *param;
    struct dict_entry *next;
    
    while (entry) {
	param = entry->param;
	dictparam_free(param);
	next = entry->next;
	free(entry);
	entry = next;
    }
}


void _dict_remove_entry(dict *pb, struct dict_entry *entry)
{
    int i;
    for (i=0;i < pb->hsize;i++) {
	if (pb->ht[i] == entry) {
	    pb->ht[i] = NULL;
	    _remove_entry_list(entry);
	    return;
	}
    }
}
		
		

/*
 * param_create creates a parameter with the given name and value. If name
 * and value are non-NULL, they are copied and placed into the new dict_param
 * struct.
 */

dict_param *dictparam_create(char *name, char *value)
{
    dict_param *new;
    new = malloc(sizeof(dict_param));
    new->name = malloc((strlen(name) + 1)*sizeof(char));
    strcpy(new->name,name);
    new->value = malloc((strlen(value) + 1)*sizeof(char));
    strcpy(new->value,value);
    return new;
}


/*
 * param_free frees a given parameter if it's non-NULL, and returns 1 if
 * p was non-NULL, and 0 if p was NULL.
 *
 * Useful for error checking dict_remove.
 */

int dictparam_free(dict_param *p)
{
    if (p) {
	if(p->name) free(p->name);
	if (p->value) free(p->value);
	free (p);
	return 1;
    } else {
	return 0;
    }
}

/*
 * dict_create creates a new dict with hash table size n.
 *
 * It returns the newly allocated dict.
 */
 
dict *dict_create(int n)
{
    dict *block;
    int i;
    
    block = malloc(sizeof(dict));
    block->hsize = n;
    block->ht = calloc(n,sizeof(struct dict_entry *));
    for (i=0;i < n;i++) {
	block->ht[i] = NULL;
    }
    return block;
}

/*
 * dict_free frees the given dict and any entries inside it.
 *
 * If you want to save anything in a dict, remove its entities with
 * dict_remove first and save the pointers you get.
 */

void dict_free(dict *pb)
{
    int i;
    for (i=0;i < pb->hsize;i++) {
	if (pb->ht[i]) {
	    struct dict_entry *entry;
	    entry = pb->ht[i];
	    _remove_entry_list(entry);
	}
    }
    free(pb->ht);
    free(pb);
}
/*
 * dict_findval finds the entry with the given name in dict pb, and
 * returns its value, otherwise returns NULL.
 */

char *dict_findval(char *name, dict *pb)
{
    struct dict_entry *found;
    found = _dict_find(name, pb);
    if (found) {
	return found->param->value;
    } else {
	return NULL;
    }
}

/*
 * dict_find finds the entry with the given name in dict pb.
 *
 * If it is successful, it returns the param block. If not, it returns
 * NULL.
 *
 * Note: This function is actually a macro which calls a separate
 * function with a hard-coded third parameter.
 */

dict *dict_find(char *name, dict *pb)
{
    if (_dict_find(name,pb)) return pb; else return NULL;
}
    

/*
 * dict_remove behaves exactly like dict_find, but removes the given
 * entry from pb.
 *
 * Note: This function is actually a macro which calls a separate
 * function with a hard-coded third parameter.
 */

dict *dict_remove(char *name, dict *pb)
{
    struct dict_entry *entry;
    entry = _dict_find(name, pb);
    if(entry)
	_dict_remove_entry(pb,entry);
    return pb;
}

    

/*
 * dict_nvinsert creates a new parameter with the given name and value
 * and inserts it into dict pb. The name and value in the parameter are
 * also newly allocated. Returns the dict_param it allocated (in case you
 * need it).
 */

dict_param *dict_nvinsert(char *name, char *value, dict *pb)
{
    struct dict_entry *entry, *newentry;
    dict_param *param;
    
    param = dictparam_create(name,value);
    newentry = malloc(sizeof(struct dict_entry));
    newentry->param = param;
    newentry->next = NULL;
    
    entry = _dict_find(name,pb);
    if (entry) {
	_insert_entry(entry,newentry);
    } else {
	_dict_insert_entry(newentry,pb);
    }
    return param;
}


/*
 * dict_pinsert inserts a dict_param into a dict.
 */

void dict_pinsert(dict_param *pp, dict *pb)
{
    struct dict_entry *entry, *newentry;
    
    newentry = malloc(sizeof(struct dict_entry));
    newentry->param = pp;
    newentry->next =NULL;
    entry = _dict_find(pp->name,pb);
    if (entry) {
	_insert_entry(entry,newentry);
    } else {
	_dict_insert_entry(newentry, pb);
    }
    return ;
}
    

