/*
 * Parameter blocks, as defined by the Netscape server API
 *
 * The Parameter block is basically dictionary containing key value pairs.
 * the keys are called 'name' and the value are called 'value'. For keys
 * that have multiple values, a linked list is built for that entry so that
 * one key may contain many 'values'.
 */
#ifndef WODICT
#define WODICT
/* a key/value pair */
typedef struct {
    char *name, *value;
} dict_param;

/* an entry which is a key value pair and a pointer to the next entry */
struct dict_entry {
    dict_param *param;
    struct dict_entry *next;
};

/* a parameter block which contains a table of entries */
typedef struct {
    int hsize;
    struct dict_entry **ht;
} dict;

/* for a definition of these functions, see netscape API */

dict *dict_create(int n);
    /* create a dict with n entries in it */
void dict_free(dict *pb);
    /* free a dict */
char *dict_findval(char *name, dict *pb);
    /* find a 'value' by 'name' in dict, return the value or NULL */
dict *dict_find(char *name, dict *pb);
    /* see if the entry by 'name' is in the dict. If yes return dict, other wise return NULL */
dict *dict_remove(char *name, dict *pb);
    /* remove an entry by that 'name' in dict. Remove the entire list if the entry contains multiple values by the same name. */
dict_param *dict_nvinsert(char *name, char *value, dict *pb);
    /* insert a new key/value pair into the dict, and return the entry (dict_param) inserted into the dict */
void dict_pinsert(dict_param *pp, dict *pb);
    /* insert a parameter entry into the plock */
dict_param *dictparam_create(char *name, char *value);
    /* create an key/value pair */
int dictparam_free(dict_param *p);
    /* free a key/value pair. Return 1 if freed, return 0 if argument is null */
#endif
