#include <stdlib.h>
#include <stdio.h>

#ifndef WIN32

#include <unistd.h>

#if (defined(NeXT_PDO) && defined(__svr4__))
#include <dirent.h>
#else /* not SOLARIS */
#include <sys/dir.h>
#endif

#else /* not WIN32 */

#include <windows.h>
#include <winsock.h>

#endif

#include <errno.h>
#include <string.h>
#include <sys/param.h>

#include <fcntl.h>

#include <sys/types.h>
#include <sys/stat.h>

#include "WODirectoryListing.h"
#include "WOAdaptorUtil.h"

#define RESPONSE_HEAD_FORMAT "<HTML><HEAD><TITLE>Index of WebObjects Applications</TITLE></HEAD>\n<BODY BGCOLOR=\"#FFFFFF\"><CENTER><H3>Could not find the application specified in the url.</H3></CENTER>\n<H4>Index of WebObjects Applications in %s</H4>\n<PRE>   Name                                Description\n<HR>\n"
#define RESPONSE_DIR_FORMAT "<A HREF=\"%s\"><IMG SRC=\"%s\" ALT=\"%s\" BORDER=0>  %s</A>"
#define RESPONSE_TAIL "<BR>\n<BR>\n<CENTER><A HREF=\"http://www.apple.com\"><IMG SRC=\"/WebObjects/icons/PoweredByWebObjects.gif\" ALT=\"Powered By WebObjects\" BORDER=0></A>\n</CENTER></BODY></HTML>"

#define BACK_ICON "/WebObjects/icons/back.gif"
#define WOAPP_ICON "/WebObjects/icons/woapp.gif"
#define DIR_ICON "/WebObjects/icons/dir.gif"
#define TEXT_ICON "/WebObjects/icons/text.gif"

#define DEFAULT_TYPE "text/html"
#define DEFAULT_STATUS_LINE "HTTP/1.1 200 OK"
#define SPACES 40

WOHTTPResponse *WOErrorResponseWithFile(const char *documentRoot, const char *htmlFile) {

    if (htmlFile && *htmlFile && documentRoot && *documentRoot) {
        char filePath[WO_MAX_PATH+1];
        struct stat statbuf;
        sprintf(filePath,"%s/%s",documentRoot,htmlFile);
        if (!stat(filePath,&statbuf)) {
            WOHTTPResponse *response = (WOHTTPResponse *) malloc(sizeof(WOHTTPResponse));
            char *statusLine = malloc((strlen(DEFAULT_STATUS_LINE) + 1)*sizeof(char));

            /* Set the status line and the headers */

            response->headers = dict_create(5);
            dict_nvinsert(CONTENT_LENGTH,"0",response->headers);
            dict_nvinsert(CONTENT_TYPE,DEFAULT_TYPE,response->headers);
            dict_nvinsert("location",(char *)htmlFile,response->headers);

            sprintf(statusLine,"HTTP/1.1 302 OK");

            /* WOResponse will free the allocated statusLine later. */
            response->statusLine = statusLine;
            return response;
        }
    }
    return NULL;
}

WOHTTPResponse *WOErrorResponse(const char *html) {

    if (html && *html) {
        WOHTTPResponse *response = (WOHTTPResponse *) malloc(sizeof(WOHTTPResponse));
        char length[WO_MAX_PATH+1];
        char *statusLine = malloc((strlen(DEFAULT_STATUS_LINE) + 1)*sizeof(char));

        response->content = malloc(strlen(html)+1);
        sprintf(response->content,html);
        /* Set the status line and the headers */

        response->headers = dict_create(5);
        sprintf(length,"%d",(int)strlen(response->content));
        dict_nvinsert(CONTENT_LENGTH,length,response->headers);
        dict_nvinsert(CONTENT_TYPE,DEFAULT_TYPE,response->headers);

        sprintf(statusLine,DEFAULT_STATUS_LINE);

        /* WOResponse will free the allocated statusLine later. */
        response->statusLine = statusLine;
        return response;
    }
    return NULL;
}

WOHTTPResponse *WODirectoryListing(const char *applicationName, const char *documentRoot, WOURLComponents *urlComponents) {
    char *p;
    struct stat statbuf;
    int len = 0;
    int i, path_info_len;
    WOHTTPResponse *response = (WOHTTPResponse *) malloc(sizeof(WOHTTPResponse));
    char *statusLine = malloc((strlen(DEFAULT_STATUS_LINE) + 1)*sizeof(char));
    char path[WO_MAX_PATH+1];
    char lastPath[WO_MAX_PATH+1];
    char name[WO_MAX_PATH+1];
    char fileName[WO_MAX_PATH+1];
    char prefix[WO_MAX_PATH+1];
    char tempPath[WO_MAX_PATH+1];
    char spaces[WO_MAX_PATH+1];
    char readme[WO_MAX_PATH+1];
    char url[WO_MAX_PATH+1];
    char wrapperPath[WO_MAX_PATH+1];

#ifdef WIN32
    WIN32_FIND_DATA file;
    HANDLE handle;
#else /* not WIN32 */

#if (defined(NeXT_PDO) && defined(__svr4__))
    struct dirent *dp;
#else
    struct direct *dp;
#endif
    DIR *dirp;
#endif

    for (i=0; i < SPACES; i++) spaces[i] = ' ';
    spaces[SPACES]='\0';
  
    logWebObjects(LOGCGI_WARNING,"-- WODirectoryListing : Creating Listing Page for : %s/WebObjects/%s",documentRoot,applicationName);
    
    strncpy(prefix,urlComponents->prefix.start,urlComponents->prefix.length);
    prefix[urlComponents->prefix.length] = '\0';
    
    strncpy(name,applicationName,strlen(applicationName)+1);
    sprintf(path,"%s/WebObjects",documentRoot);
    strncpy(lastPath,path,strlen(path)+1);

    while (strlen(name)) {
        
        sprintf(tempPath,"%s/",path);
        p = strchr(name,'/');
        if (p) {
            p[0] = '\0';
        }
        strcat(tempPath,name);
        if (stat(tempPath,&statbuf) || !(statbuf.st_mode & S_IFDIR) || strstr(name, WOADAPTOR_APP_EXTENSION) || strstr(name, WOADAPTOR_COMPONENT_EXTENSION)) {
            name[0] = '\0';
            break;
        }
        strncpy(lastPath,path,strlen(path)+1);
        strncpy(path,tempPath,strlen(tempPath)+1);
        if (p) {
            p++;
            strncpy(name,p,strlen(p)+1);
        } else {
            name[0] = '\0';
        }
    }

    /* Clean up lastPath & path */
        for (len = strlen(lastPath); len < WO_MAX_PATH +1; len++) {
            lastPath[len] = '\0';
        }
            
        if (lastPath[strlen(lastPath)-1] == '/') lastPath[strlen(lastPath)-1] = '\0';
        for (len = strlen(path); len < WO_MAX_PATH+1; len++) {
            path[len] = '\0';
        }

            if (path[strlen(path)-1] == '/') path[strlen(path)-1] = '\0';
    
/* Now do a first pass on the directory path to compute the approximate size of the response */
    
    len += strlen(RESPONSE_HEAD_FORMAT)+strlen(RESPONSE_TAIL);
    len += strlen(path);

#ifdef WIN32
    sprintf(tempPath,"%s/*",path);
    handle = FindFirstFileA(tempPath, &file);
    if (INVALID_HANDLE_VALUE == handle) {
        return NULL;
    }
    do {
        len += WO_MAX_PATH*3*sizeof(char);
    } while (FindNextFileA(handle, &file));
#else /* not WIN32 */
    dirp = opendir(path);
    if (!dirp) {
	logWebObjects(LOGCGI_WARNING,"-- WODirectoryListing : Cannot find directory at path :%s . Error \"%s\" occured.\n",path,strerror(errno));
	return NULL;
    }
    for (dp = readdir(dirp); dp != NULL; dp = readdir(dirp)) {
        len += WO_MAX_PATH*3*sizeof(char);
    }
    closedir(dirp);
#endif
/* Now do a second pass on the directory path to fill up the response */

    response->content = malloc(len+1);
    sprintf(response->content,RESPONSE_HEAD_FORMAT,path);

#ifdef WIN32
    sprintf(tempPath,"%s/*",path);
    handle = FindFirstFileA(tempPath, &file);
    if (INVALID_HANDLE_VALUE == handle) {
        return NULL;
    }
    do {
	strncpy(fileName,file.cFileName,strlen(file.cFileName)+1);
#else    
    dirp = opendir(path);
    if (!dirp) {
	logWebObjects(LOGCGI_WARNING,"-- WODirectoryListing : Cannot find directory at path :%s . Error \"%s\" occured.\n",path,strerror(errno));
	return NULL;
    }
    for (dp = readdir(dirp); dp != NULL; dp = readdir(dirp)) {
        strncpy(fileName,dp->d_name,strlen(dp->d_name)+1);
#endif
        path_info_len = strlen(documentRoot)+strlen("/WebObjects");
        sprintf(tempPath,"%s/%s",path,fileName);

        if (!stat(tempPath,&statbuf) && (statbuf.st_mode & S_IFDIR)){
            char *p;

/* look for a ReadMe.html file */
            sprintf(readme,"%s/ReadMe.html",tempPath);
            if (!stat(readme,&statbuf)) {
		readme[path_info_len] = '\0';
                p = readme+(path_info_len+1)*sizeof(char);
		strncpy(url,p,strlen(p)+1);
                sprintf(readme,"<A HREF=\"/WebObjects/%s\"><IMG SRC=\"%s\" ALT=\"ReadMe.html\" BORDER=0></A>\n",url,TEXT_ICON);
            } else {
/* look for a Resources/ReadMe.html file */
                sprintf(readme,"%s/Resources/ReadMe.html",tempPath);
                if (!stat(readme,&statbuf)) {
                    readme[path_info_len] = '\0';
                    p = readme+(path_info_len+1)*sizeof(char);
                    strncpy(url,p,strlen(p)+1);
                    sprintf(readme,"<A HREF=\"/WebObjects/%s\"><IMG SRC=\"%s\" ALT=\"ReadMe.html\" BORDER=0></A>\n",url,TEXT_ICON);
                } else {
/* look for a WebServerResources/ReadMe.html file */
                    sprintf(readme,"%s/WebServerResources/ReadMe.html",tempPath);
                    if (!stat(readme,&statbuf)) {
                        readme[path_info_len] = '\0';
                        p = readme+(path_info_len+1)*sizeof(char);
                        strncpy(url,p,strlen(p)+1);
                        sprintf(readme,"<A HREF=\"/WebObjects/%s\"><IMG SRC=\"%s\" ALT=\"ReadMe.html\" BORDER=0></A>\n",url,TEXT_ICON);
                    } else {
                        readme[0]='\0';
                    }
                }
            }


            if (strcmp(fileName,".")) {
/* Check if the directory is .. */
                if (strstr(fileName,"..")) {
                    if (strlen(path) > path_info_len+1) {
                        lastPath[path_info_len] = '\0';
                        p = lastPath + (path_info_len + 1)*sizeof(char);
                        sprintf(name,"%s/%s",prefix,p);
                        sprintf(response->content+strlen(response->content)*sizeof(char), RESPONSE_DIR_FORMAT,name, BACK_ICON, "[BACK]", "Parent Directory");
                        sprintf(response->content+strlen(response->content)*sizeof(char), "\n");
                    }
                } else {

/* check if the directory is a WOApp OR
 * Check if the directory under this directory is a .woa .
 * As of WOF3.5, we should autostart an application if its project
 * directory contains a .woa with the same name.
 */
                    sprintf(wrapperPath,"%s/%s%s",tempPath, fileName, WOADAPTOR_APP_EXTENSION);

                    if (strstr(fileName, WOADAPTOR_APP_EXTENSION) || (!stat(wrapperPath,&statbuf) && (statbuf.st_mode & S_IFDIR))) {
                        char *q;
                        tempPath[path_info_len] = '\0';
                        p = tempPath + (path_info_len + 1)*sizeof(char);
                        sprintf(name,"%s/%s",prefix,p);
/* remove the .woa terminaison if needed */
                        q = strstr(p, WOADAPTOR_APP_EXTENSION);
                        if (q) q[0]='\0';		

                        sprintf(response->content +strlen(response->content)*sizeof(char), RESPONSE_DIR_FORMAT, name, WOAPP_ICON, "[WOAPP]", p);
                        strncat(response->content,spaces,SPACES-strlen(p));
                        sprintf( response->content+strlen(response->content)*sizeof(char), strlen(readme) ? readme : "\n");
                    } else {
/* Check if the directory is a WO page */
                        if (!strstr(fileName, WOADAPTOR_COMPONENT_EXTENSION)) {
                            tempPath[path_info_len] = '\0';
                            p = tempPath + (path_info_len + 1)*sizeof(char);

/* Special case for the Documentation folder */
                            if (!strstr(fileName,"Documentation")) {
                                sprintf(name,"%s/%s",prefix,p);
                            } else {
                                sprintf(name,"/WebObjects/%s",p);
                            }
                            sprintf( response->content+strlen(response->content)*sizeof(char), RESPONSE_DIR_FORMAT,name, DIR_ICON, "[DIR]",p);
                            strncat(response->content,spaces,SPACES-strlen(p));
                            sprintf( response->content+strlen(response->content)*sizeof(char), strlen(readme) ? readme : "\n");
                        }
                    }
                }
            }
        }
    } 
#ifdef WIN32
      while (FindNextFileA(handle, &file));
#else
    closedir(dirp);
#endif

    strcat(response->content,RESPONSE_TAIL);

/* Set the status line and the headers */
    
    response->headers = dict_create(20);
    sprintf(name,"%d",(int)strlen(response->content));
    dict_nvinsert(CONTENT_LENGTH,name,response->headers);
    dict_nvinsert(CONTENT_TYPE,DEFAULT_TYPE,response->headers);

    sprintf(statusLine,DEFAULT_STATUS_LINE);
    
/* WOResponse will free the allocated statusLine later. */
    response->statusLine = statusLine;

    return response;
}

