/*
    WOHTTPProtocol.m
    (c) Copyright 1996 NeXT Software Inc.
    All Rights Reserved.
*/

#include <stdio.h>
#include <stdlib.h>
#ifndef WIN32
#include <unistd.h>
#else
#include <winsock.h>
#endif
#include <string.h>
#include <ctype.h>
#include "WOHTTPProtocol.h"
#include "WOAdaptorUtil.h"	/* utility routines */

int WOHTTPRequestSend(FILE *fp, WOHTTPRequest *rq) {
    struct dict_entry *dictEntry;
	char hugeBuffer[HUGE_BUFFER];
    char lineBuffer[HUGE_STRING_LENGTH];
    dict_param *dictParam;
    char *length = NULL;
    long contentLength=0L;
    int i;

    /* write the first request line */
    
    length = dict_findval("content-length",rq->headers);
    if (length) {
	contentLength = atol(length);
	if (contentLength && !rq->content) {
#ifndef WIN32
	    logWebObjects(LOGCGI_WARNING,"-- WOHTTPRequestSend: Error : content-length non nul but no content");
	    return -1;
#else
	    char lenBuffer[64];
	    contentLength=0L;
	    sprintf(lenBuffer,"%ld",contentLength);
	    dict_nvinsert("content-length",lenBuffer,rq->headers);
#endif	    
	    }
    }



    if (rq->requestLine) {
#ifdef NORMALIZE_GET
	char *p, *q, *gcontent=NULL;
#else
	char *gcontent=NULL;
#endif
	int form=0;

	strcpy(hugeBuffer,"");
	strncpy(lineBuffer,rq->requestLine, HUGE_STRING_LENGTH -1 );
	lineBuffer[HUGE_STRING_LENGTH - 1] = '\0';

/* Now we no longer normalize the GET requests into POST on the adaptor side. */
#ifdef NORMALIZE_GET
	if (!strncmp(lineBuffer,"GET",3*sizeof(char)) && (p = strchr(lineBuffer+4,'?'))) {
	    form = 1;
	    *p = '\0';
	    q = strchr(p+1,' ');
	    if (!q) {
		q = ++p;
	    } else {
		*q++ = '\0';
		gcontent = ++p;
		contentLength = (long)strlen(gcontent);
		dict_nvinsert("content-type","application/x-www-form-urlencoded",rq->headers);
	    }
	    strcat(hugeBuffer,"POST");
	    strcat(hugeBuffer,lineBuffer+3);
	    strcat(hugeBuffer," ");
	    strcat(hugeBuffer,q);
	    logWebObjects(LOGCGI_INFO,"-- WOHTTPRequestSend: requestLine: POST%s %s",lineBuffer+3,q);
	} else {
	    strcat(hugeBuffer,rq->requestLine); /* first line */
	}
#else
        strcat(hugeBuffer,rq->requestLine); /* first line */
#endif

	strcat(hugeBuffer,"\n");
#ifndef WIN32
	fputs(hugeBuffer,fp);
	fflush(fp);
#else
	send ((int)fp, hugeBuffer, strlen (hugeBuffer), 0);
	WSAGetLastError();
#endif

/* If content-length is not yet in the headers, or we changed from GET to POST, add the content-length */
	if (!length || form) {
	    char lenBuffer[64];
	    sprintf(lenBuffer,"%ld",contentLength);
	    dict_nvinsert("content-length",lenBuffer,rq->headers);
	    }
	for (i=0;i < rq->headers->hsize;i++) { /* headers */
	    dictEntry = rq->headers->ht[i];
	    while (dictEntry) {
		dictParam = dictEntry->param;
		if (dictParam && dictParam->name && dictParam->value) {
		    hugeBuffer[0] = '\0';
		    strcat(hugeBuffer,dictParam->name);
		    strcat(hugeBuffer,": ");
		    strcat(hugeBuffer,dictParam->value);
		    strcat(hugeBuffer,"\n");
		    logWebObjects(LOGCGI_INFO,"-- WOHTTPRequestSend: %s: %s",dictParam->name, dictParam->value); 
#ifndef WIN32
		    fputs(hugeBuffer,fp);
		    fflush(fp);
#else
		    send ((int)fp, hugeBuffer, strlen (hugeBuffer), 0);
		    WSAGetLastError();
#endif
		}
		dictEntry = dictEntry->next;
		if (!dictEntry) break;
	    }
	}
	hugeBuffer[0] = '\0';
	strcat(hugeBuffer,"\n");

#ifndef WIN32
	fputs(hugeBuffer,fp);
	fflush(fp);
#else
	send ((int)fp, hugeBuffer, strlen (hugeBuffer), 0);
	WSAGetLastError();
#endif

	if (contentLength) {
	    if (form) {
#ifndef WIN32
	    fwrite(gcontent,contentLength,1,fp);
#else
		send ((int)fp, gcontent, contentLength, 0);
#endif
	} else {
#ifndef WIN32
		fwrite(rq->content, contentLength,1,fp);
#else
		send ((int)fp, rq->content, contentLength, 0);
#endif
	    }
	}
#ifndef WIN32
	fflush(fp);
#endif
	return 0;
    } else {
	logWebObjects(LOGCGI_WARNING,"-- WOHTTPRequestSend: Error : No request line !");
	return -1;
    }
}

WOHTTPResponse * WOHTTPResponseReceive(FILE *fp) {
    char lineBuffer[HUGE_STRING_LENGTH];
    char *value;
    WOHTTPResponse *rp;
    char *len;
    int datalen;
	    
    logWebObjects(LOGCGI_INFO,"-- WOHTTPResponseReceive: Receiveresponse.");

    if (!getLine(lineBuffer,sizeof(lineBuffer),fp)) {
	return NULL;
    }
    
    logWebObjects(LOGCGI_INFO,"-- WOHTTPResponseReceive: status %s",lineBuffer);

    rp = (WOHTTPResponse *)malloc(sizeof(WOHTTPResponse));
    rp->statusLine=NULL;
    rp->headers=NULL;
    rp->content=NULL;
    
    rp->statusLine = malloc((strlen(lineBuffer) + 1) * sizeof(char));
    strcpy(rp->statusLine,lineBuffer);
    
    rp->headers = dict_create(30);
    
    while (getLine(lineBuffer,sizeof(lineBuffer),fp)) {
	if (!lineBuffer[0]) break;	/* possibly content to follow */
	
	if (!(value = strchr(lineBuffer,':'))) continue;
	
	*value++ = '\0';
	while (isspace(*value)) ++value;
	
	/* now merge the key values */
	dict_nvinsert(lowerCaseString(lineBuffer),value,rp->headers);
	logWebObjects(LOGCGI_INFO,"-- WOHTTPResponseReceive: %s: %s",lineBuffer,value);

    }
    len = dict_findval("content-length",rp->headers);

    if (len) {
	datalen = atoi(len);
	logWebObjects(LOGCGI_INFO,"-- WOHTTPResponseReceive: content length: %d",datalen);
	rp->content = malloc(datalen);
#ifdef WIN32
	{
		int totalAmountReceived = 0;
		int lastError;
		while (totalAmountReceived < datalen)
		{
			lastError = recv ((int)fp, rp->content + totalAmountReceived, datalen - totalAmountReceived, 0);
			if (lastError != SOCKET_ERROR)
				totalAmountReceived += lastError;
			else 
				break;
		}
	}
#else
	fread(rp->content,datalen,1,fp);
#endif
    }
    else {
 	logWebObjects(LOGCGI_INFO,"-- WOHTTPResponseReceive: no content length header received.");
    }

    return rp;
}

void WO_request_free(WOHTTPRequest *request)
{
    if (request->requestLine) {
	free(request->requestLine);
	}
    if (request->headers) {
	dict_free(request->headers);
	}
    if (request->content) {
	free(request->content);
	}
    
    free(request);
    return;
}

void WO_response_free(WOHTTPResponse *response)
{
    if (response->statusLine) {
	free(response->statusLine);
	}
    if (response->headers) {
	dict_free(response->headers);
	}
    if (response->content) {
	free(response->content);
	}
    
    free(response);
    return;
}

   
