#include <stdlib.h>
#ifndef WIN32
#include <unistd.h>
#else
#include <windows.h>
#endif
#include <fcntl.h>
#include <sys/stat.h>
#include <errno.h>
#include <string.h>
#if (defined(NeXT_PDO) && defined(__alpha__)) || defined(__sgi)
#include <sys/file.h>
#endif

#ifdef NeXT
#include <libc.h>
#endif

#include "WOAdaptorUtil.h"
#include "WOLock.h"

#if defined(NeXT_PDO) && defined(__alpha__)
extern int flock(int, int);	/* cant find this anywhere! */
#endif

/* Opening, creating if needed and locking a file.
 * If file is already locked, we stall
 * until the other process unlocks...
 */
 
int WOLockOrWait(const char *tmpfile) {
    int fd;
    
#ifdef WIN32
	SECURITY_ATTRIBUTES sa;

	sa.nLength = sizeof (SECURITY_ATTRIBUTES);
	sa.lpSecurityDescriptor = NULL;
	sa.bInheritHandle = FALSE;

    fd = (int)CreateFile(tmpfile,GENERIC_READ | GENERIC_WRITE, FILE_SHARE_READ | FILE_SHARE_WRITE, &sa, OPEN_ALWAYS, FILE_ATTRIBUTE_NORMAL, NULL);
	if ((HANDLE)fd != INVALID_HANDLE_VALUE)
	{
		while (!LockFile ((HANDLE)fd, 1, 0, 1, 0))
		{
			Sleep (1000);
		}
		return fd;
	}
#else /* !WIN32 */
    umask(0000);
    fd = open(tmpfile,O_CREAT | O_WRONLY, 0666);

    if (fd >=  0) {
	close(fd);
    }
    else {
        logWebObjects(LOGCGI_WARNING,"-- WOLockOrWait : Error \"%s\" occured while opening %s",strerror(errno),tmpfile);
	umask(0022);
	return -1;
    }
    umask(0022);		/* reset user mask */

    fd = open(tmpfile,O_RDWR) ;
    if (fd > 0) {
        int r;
 	logWebObjects(LOGCGI_INFO,"-- WOLockOrWait: Opened %s",tmpfile);

/*
    	struct flock lock;
	lock.l_type = F_WRLCK;
	lock.l_start = 0;
	lock.l_whence = 0;
	lock.l_len = 0; 		
	r = fcntl(fd,F_SETLKW,&lock);
*/
#if ((NeXT_PDO && (__svr4__ || hpux)) || __hpux)
	r = lockf(fd,F_LOCK,(long) 0);
#else
	r = flock(fd,LOCK_EX);
#endif
	if (!r) {
	    return fd;
	} 
	else {
	    
	    logWebObjects(LOGCGI_WARNING,"-- WOLockOrWait : Error \"%s\" occured while locking %s",strerror(errno),tmpfile);
	    close(fd);
	    return -1;
	}
    } 
#endif /* WIN32 */
    logWebObjects(LOGCGI_WARNING,"-- WOLockOrWait : Error \"%s\" occured while opening %s",strerror(errno),tmpfile);

    return -1;
}

/* Opening, creating if needed and locking a file.
 * If file is already locked, we return an error (-1).
 */

int WOLockOrFail(const char *tmpfile) {
    int fd;

#ifdef WIN32
	SECURITY_ATTRIBUTES sa;

	sa.nLength = sizeof (SECURITY_ATTRIBUTES);
	sa.lpSecurityDescriptor = NULL;
	sa.bInheritHandle = FALSE;

    fd = (int)CreateFile(tmpfile,GENERIC_READ | GENERIC_WRITE, 0, &sa, OPEN_ALWAYS, FILE_ATTRIBUTE_NORMAL, NULL);
	if ((HANDLE)fd != INVALID_HANDLE_VALUE)
	{
		if (LockFile ((HANDLE)fd, 1, 0, 1, 0))
		{
			return fd;
		}
		CloseHandle ((HANDLE)fd);
	}
#else /* !WIN32 */
/* set the user mask to nothing, so that we can write from a 'nobody' user */
    umask(0000);

    fd = open(tmpfile,O_CREAT | O_WRONLY, 0666);
    if (fd >=  0) {
	close(fd);
    }
    else {
        logWebObjects(LOGCGI_WARNING,"-- WOLockOrFail : Error \"%s\" occured while opening %s",strerror(errno),tmpfile);
	umask(0022);
	return -1;
    }
    fd = open(tmpfile,O_RDWR) ;
    umask(0022);		/* reset user mask */

    if (fd > 0) {
    	int r;
    	logWebObjects(LOGCGI_INFO,"-- WOLockOrFail: Opened %s",tmpfile);

/*
        struct flock lock;
	lock.l_type = F_WRLCK;
	lock.l_start = 0;
	lock.l_whence = 0;
	lock.l_len = 0; 
	r = fcntl(fd,F_SETLK,&lock);
*/

#if ((NeXT_PDO && (__svr4__ || hpux)) || __hpux)
	r = lockf(fd,F_TLOCK,0);
#else
	r = flock(fd,LOCK_EX | LOCK_NB);
#endif	
	if (!r) {
	    return fd;
	} 
	logWebObjects(LOGCGI_INFO,"-- WOLockOrFail: Error \"%s\" occured while locking %s",strerror(errno),tmpfile);	
    } 
    else {
	logWebObjects(LOGCGI_WARNING,"-- WOLockOrFail : Error \"%s\" occured while opening %s",strerror(errno),tmpfile);
    }
#endif /* WIN32 */
    return -1;
}

int WOUnlock(int fd) {
    int r;
#ifdef WIN32
	// Unlock File returns TRUE if successful 
	r = !(UnlockFile((HANDLE)fd, 1, 0, 1, 0));
	CloseHandle ((HANDLE)fd);
#else /* !WIN32 */
/*
    struct flock lock;
    lock.l_type = F_UNLCK;
    lock.l_start = 0;
    lock.l_whence = 0;
    lock.l_len = 0; 		    	
    r = fcntl(fd,F_SETLK,&lock);
*/

#if ((NeXT_PDO && (__svr4__ || hpux)) || __hpux)
    r = lockf(fd,F_ULOCK,0);
#else
    r = flock(fd,LOCK_UN);
#endif
#endif /* WIN32 */
    if (r) {
	logWebObjects(LOGCGI_WARNING,"-- WOUnlock : error \"%s\" occured while unlocking a file",strerror(errno));
	return -1;
    }
    logWebObjects(LOGCGI_INFO,"-- WOUnlock: Unlocked file");
    return 0;
}
