#include <stdlib.h>
#ifndef WIN32
#include <unistd.h>
#else
#include <windows.h>
#include <winsock.h>
#endif
#include <fcntl.h>
#include <sys/stat.h>
#include <errno.h>
#include <string.h>

#ifdef NeXT
#include <libc.h>
#endif

#if (NeXT_PDO && __svr4__ )
int gethostname(char *, int);
#endif

#include "WOSocket.h"
#include "WOServerTmpConfig.h"

/* returns a socket descriptor and for status
 *	0 for not found
 *	1 for found
 * 	-1 for error occured
 */

int WOGetSocketFromPortInTMPConfig(FILE *fp, const char *appName, const char *hostName,int pid, int *status) {
    struct app_table record;
    int r = 1;
    *status = -1;

    logWebObjects(LOGCGI_INFO,"-- WOGetSocketFromPortInTMPConfig : Looking for app %s (no instance) in config",appName);
    rewind(fp);
    while (r > 0) {
	r = fread(&record,sizeof(struct app_table),1,fp);
	if (!r) {
	    logWebObjects(LOGCGI_INFO,"-- WOGetSocketFromPortInTMPConfig : App %s not found in config" , appName);
	    *status = 0;
	    return 0;
	    }
	if (r < 0) {
	    logWebObjects(LOGCGI_WARNING,"-- WOGetSocketFromPortInTMPConfig : Error \"%s\" occured while reading %s info from temporary configuration file",strerror(errno),appName);
	    *status = -1;
	    return 0;
	}
	if ((record.pid != -1) && (record.port != -1)) {
	    logWebObjects(LOGCGI_INFO,"-- WOServerTmpConfig : read \"%s\" %d %d",record.appName, record.port, record.pid);
	}
	    
	if ((appName) && (record.appName) && (!strcmp(appName,record.appName)) && (record.port != -1) && (((pid == DO_NOT_CHECK_PID) && (record.pid != -1)) || ((pid != DO_NOT_CHECK_PID) && (record.pid == pid)))) {

/* This part was before implemented as a kill(record.pid,0)
 * but problems of ownership occured.
 * now we try to connect to the port specified for this application
 * if the connection refused error comes back, we'll fail,
 * and erase this port from the tmp file (assuming the process is dead
 */
	    int sd ;

	    sd = WOSocketConnectTo(hostName,record.port, 100);

/* A second attempt on NeXT and Solaris with localhost
 * in the case when gethostbyname() failed... 
 * sd == -2 checks for this special case.
 */
#ifndef	WIN32
	    if (sd == -2) {
		sd = WOSocketConnectTo("localhost",record.port,100);
	    }
#endif		
	    if (sd < 0) {
		logWebObjects(LOGCGI_INFO,"-- WOGetSocketFromPortInTMPConfig : App %s matching but error occured while checking port %d. Deleting App from tmp config file.",appName, record.port);

/* This is the case when something bad happened on the application side,
 * the application is up and running (matched pid)
 * but the port did not respond. We log an error. We can not exit with
 * an error as in some race conditions, the port and pid may be listed
 * by the App in the tmp config BEFORE that the app is ready for connection.
 */
		if ((pid != DO_NOT_CHECK_PID) && (record.pid == pid)) {
		    logWebObjects(LOGCGI_WARNING,"-- WOGetSocketFromPortInTMPConfig : Process (%d) was matching forked process. App (%s) may have hang at start up.", pid, appName);
		}
		fseek(fp,sizeof(struct app_table) * (-1), SEEK_CUR);
		record.port = -1;
		record.pid = -1;
		if (!fwrite(&record,sizeof(struct app_table),1,fp)) {
		    logWebObjects(LOGCGI_WARNING," -- WOGetSocketFromPortInTMPConfig : Error \"%s\" occured while deleting %s from config file", strerror(errno), appName);
		    *status = -1;
		    return 0;
		}
		logWebObjects(LOGCGI_INFO,"-- WOGetSocketFromPortInTMPConfig : Returning App not found.");
		*status = 0;
		return 0;
	    }	    
	     else {
		logWebObjects(LOGCGI_INFO,"-- WOGetSocketFromPortInTMPConfig : Matched app %s and port %d is listening !" , record.appName, record.port);
		*status = 1;
		return sd;
	    }
	}
    }
    return 0;
}

