#include <stdio.h>
#include <stdlib.h>
#ifndef WIN32
#include <unistd.h>
#include <sys/socket.h>
#include <netinet/in.h>
#include <arpa/inet.h>
#include <netdb.h>
#else
#include <winsock.h>
#endif
#include <errno.h>
#include <sys/types.h>
#include <string.h>
#include <signal.h>
#include <setjmp.h>
#ifdef NeXT
#include <libc.h>
#endif

#include "WOAdaptorUtil.h"

#ifdef WIN32
unsigned int alarm (unsigned seconds)
{
	return 0;
}
#endif 

static jmp_buf timebuf;

static void timedout(int sig)
{
    alarm(0);
    longjmp(timebuf,sig);
}
/* WOSocketConnectTo() returns < 0 if fails 
 * In particular, -2 if gethostbyname fails
 */
 
int WOSocketConnectTo(const char *host, int port, unsigned int timeout)
{    
    int	sockfd=0;
    int returnSocket = -1;
    struct in_addr *inptr;
    struct sockaddr_in	serv_addr;
    struct hostent *hostptr;
#ifdef WIN32
    int reuse = 1;
#endif

    hostptr = gethostbyname( (char *) host);
    
    if (!hostptr) {
	logWebObjects(LOGCGI_WARNING,"-- WOSocket : gethostbyname() returns no host");
	return -2;
	}
    if (hostptr->h_addrtype != AF_INET) {
	logWebObjects(LOGCGI_WARNING,"-- WOSocket : wrong address type in hostptr");
	return -2;
	}
    
    inptr = (struct in_addr *) hostptr->h_addr_list[0];

#ifdef NeXT_PDO
    memset((char *) &serv_addr,0,sizeof(serv_addr));
#else
    bzero((char *) &serv_addr,sizeof(serv_addr));
#endif



    serv_addr.sin_family = AF_INET;
    serv_addr.sin_addr.s_addr = inptr->s_addr;
    serv_addr.sin_port	= htons(port);
	
    if ((sockfd = socket(AF_INET,SOCK_STREAM,0)) < 0) {
	logWebObjects(LOGCGI_WARNING,"-- WOSocket : socket() raised with \"%s\"",strerror(errno));
	return -1;
	}

#ifdef WIN32
    if (setsockopt(sockfd, SOL_SOCKET, SO_REUSEADDR, (char *)&reuse, sizeof(reuse))) {
    	logWebObjects(LOGCGI_WARNING,"-- Transaction : Could not enable socket reuse, continuing...");
    }
#endif
    
    if (setjmp(timebuf)==0) {
#ifndef WIN32
        signal(SIGALRM,timedout);
#endif 
	alarm(timeout);
	if (connect(sockfd,(struct sockaddr *) &serv_addr, sizeof(serv_addr)) < 0) {
	    logWebObjects(LOGCGI_WARNING,"-- WOSocket : connect() raised with \"%s\"",strerror(errno));
#ifdef WIN32
	    closesocket(sockfd);
#else
	    close(sockfd);
#endif
	    returnSocket = -1;
	} else {
	    returnSocket=sockfd;
	}
	alarm(0);
    } else {
#ifdef WIN32
	if (sockfd) closesocket(sockfd);
#else
	if (sockfd) close(sockfd);
#endif
	logWebObjects(LOGCGI_WARNING,"-- WOSocket : timed out (%d sec)",timedout);
	returnSocket = -1;
    }	
    return returnSocket;
}



