#define WODIRECTORYLISTING
#include <stdio.h>
#include <stdlib.h>
#ifndef WIN32
#include <unistd.h>
#include <sys/socket.h>
#else
#include <direct.h>
#include <io.h>
#include <winsock.h>
#include <windows.h>
#include <stdlib.h>
#include <string.h>
#define sleep(a) Sleep(a*1000)
#define fdopen _fdopen	/* This not getting done in stdio.h on NT */
#endif
#include <string.h>
#include <sys/param.h>
#include <fcntl.h>
#include <sys/types.h>
#include <errno.h>
#include <sys/stat.h>
#ifdef NeXT
#include <libc.h>
#include <errno.h>
#endif

#if (NeXT_PDO && __svr4__)
#include <setjmp.h>
#include <signal.h>
extern int gethostname(char *, int);
extern unsigned int sleep(unsigned int a);
#endif

#include "WOHTTPProtocol.h"
#include "WOServerAdaptor.h"
#include "WOAdaptorUtil.h"
#include "WOServerTmpConfig.h"
#include "WOLock.h"
#include "WOConfigFile.h"
#include "WOSocket.h"
#include "WOURLCUtilities.h"
#include "WOTransaction.h"
#include "WODirectoryListing.h"
#include "WORecording.h"

#define WOSERVER "WODefaultApp"
#define WEBOBJECTSAPPS "WOApps"
#define EXECUTABLES "Executables"
#define WEBOBJECTS "WebObjects"
#if (NeXT_PDO && __svr4__)
#define EXECROOT "/NextLibrary"
#elif (WIN32)
#define EXECROOT WOExecRoot()
#define EXECROOT_DIR "/NextLibrary"
#else
#define EXECROOT "/NextLibrary"
#endif
#define TMPAUTOSTARTFILE "/tmp/WebObjects.autostart"

#ifdef WIN32
#define TMPAUTOSTARTFILENAME "WebObjects.autostart"

__declspec(dllimport) const char* WOReadKeyFromConfiguration (char* keyName, char *buffer, int bufferSize);

/* The next 3 functions malloc some strings, but we don't want to free them :
- in the cgi case, they will be lost when the process goes down,
- in the .dll cases, the strings should stay valid as they are static, and may be needed
for each new connection.
*/

const char *WOExecRoot()
{
	static char* execRoot = NULL;

	if (!execRoot)
	{
		char tempDir[WO_MAX_PATH+1];
		
                WOReadKeyFromConfiguration ("NEXT_ROOT", tempDir, WO_MAX_PATH);

		execRoot = (char*)malloc ((strlen (tempDir) + strlen (EXECROOT_DIR) + 2) * sizeof (char));
		sprintf (execRoot, "%s/%s", tempDir, EXECROOT_DIR);
	}

	return execRoot;
}

const char *WOTmpAutoStartFile()
{
	static char* autoStartFileName = NULL;

	if (!autoStartFileName)
	{
            char tempDir[WO_MAX_PATH+1];

            WOReadKeyFromConfiguration ("TEMP", tempDir, WO_MAX_PATH);

            autoStartFileName = (char*)malloc ((strlen (tempDir) + strlen (TMPAUTOSTARTFILENAME) + 2) * sizeof (char));
            sprintf (autoStartFileName, "%s/%s", tempDir, TMPAUTOSTARTFILENAME);
	}

	return autoStartFileName;
}

const char *WOTmpFile()
{
	static char* tmpFileName = NULL;

	if (!tmpFileName)
	{
            char tempDir[WO_MAX_PATH+1];

            WOReadKeyFromConfiguration ("TEMP", tempDir, WO_MAX_PATH);

            tmpFileName = (char*)malloc ((strlen (tempDir) + strlen (TMPFILENAME) + 2) * sizeof (char));
            sprintf (tmpFileName, "%s/%s", tempDir, TMPFILENAME);
	}

	return tmpFileName;
}
#endif 

#if (NeXT_PDO && __svr4__)
static jmp_buf env_alrm;

static void sig_alrm(int signo) {
    longjmp(env_alrm, 1);
}

unsigned int sleep2(unsigned int nsecs) {
    if (signal(SIGALRM, sig_alrm) == SIG_ERR)
	return (nsecs);
    if (setjmp(env_alrm) == 0) {
	alarm(nsecs);
	pause();
    }
    return (alarm(0));
}
#endif

/* This function checks if the config file exists.
   It is used to decide if it is allowed to return a directory listing
   so if WODIRECTORYLISTING is not defined, it will allways return 1.
*/

int configFileExists() {
#ifdef WODIRECTORYLISTING
    char configFile[WO_MAX_PATH+1];
    struct stat statbuf;
    sprintf(configFile, "%s/%s", (char *)EXECROOT, CONFIG_FILE_PATH);
    if (!stat(configFile,&statbuf) && (statbuf.st_uid == 0)) {
	return 1;
    }
    if (!stat(CONFIG_FILE,&statbuf) && (statbuf.st_uid == 0)) {
	return 1;
    }
    return 0;
#else
    return 1;
#endif
}

static void WORequestFromURLComponents(WOHTTPRequest *request, const char *appName, int appNumber, const char *hostName, WOURLComponents *urlComponents, const char *httpVersionString) {

    char appNumberString[sizeof(appNumber) * 3 + 1];
    int versionLength = httpVersionString ? strlen(httpVersionString) : 0;

    if (appNumber > 0) {
        sprintf(appNumberString, "%d", appNumber);
    } else {
        appNumberString[0] = '\0';
    }
    urlComponents->applicationName.start = appName;
    urlComponents->applicationName.length = strlen(appName);
    urlComponents->applicationNumber.start = appNumberString;
    urlComponents->applicationNumber.length = strlen(appNumberString);
    urlComponents->applicationHost.start = hostName;
    urlComponents->applicationHost.length = strlen(hostName);

    request->requestLine = malloc(5 + 1 + WOSizeURL(urlComponents) + 1 + versionLength + 1);
    if (request->content) {
        sprintf(request->requestLine, "POST ");
    }
    else {
        sprintf(request->requestLine,"GET ");
    }
    WOComposeURL(request->requestLine + strlen(request->requestLine), urlComponents);
    sprintf(request->requestLine + strlen(request->requestLine), " %s", httpVersionString);
}

static WOHTTPResponse *Transaction(const char *appName, int applicationNumber, int sd, const char *hostName, int timeout, WOHTTPRequest *request, WOURLComponents *urlComponents, const char *httpVersionString, const char *recordingDirPath) {
#ifndef WIN32
    FILE *rfp=NULL, *wfp=NULL;
#endif
    WOHTTPResponse *r=NULL;
    int stat;

    if (!hostName) {
	logWebObjects(LOGCGI_WARNING,"-- Transaction : Error, NULL host passed.");
#ifdef WIN32
	closesocket(sd);
#else
	close(sd);
#endif
	return NULL;
	}
/* first reformat the requestLine to add the instance and the hostName if needed */
    WORequestFromURLComponents(request,appName,applicationNumber,hostName,urlComponents,httpVersionString);
    logWebObjects(LOGCGI_INFO,"-- Transaction : new request line is: %s",request->requestLine); 

/* Do not worry, this will only do anything in the case of the WebObjects-Recording adaptor */
        if (recordRequest(recordingDirPath, request) < 0) {
            logWebObjects(LOGCGI_WARNING,"-- Transaction : failed recording request.");
	}

#ifdef WIN32
	if (sd > 0) {
	    stat = WOHTTPRequestSend((FILE*)sd,request);
#else
	if ((sd > 0) && ((wfp = fdopen(sd,"w")))) {
	    stat = WOHTTPRequestSend(wfp,request);
#endif
#ifdef WIN32
	    if (!stat && ( r = WOHTTPResponseReceive((FILE*)sd))) {
#else
	if (!stat && (rfp = fdopen(sd,"r")) && ( r = WOHTTPResponseReceive(rfp))) {
#endif
	    logWebObjects(LOGCGI_INFO,"-- Transaction : transaction successful");
/* Do not worry, this will only do anything in the case of the WebObjects-Recording adaptor */
            if (recordResponse(recordingDirPath, r) < 0) {
                logWebObjects(LOGCGI_WARNING,"-- Transaction : failed recording response.");
            }
	} else {
	    logWebObjects(LOGCGI_WARNING,"-- Transaction : cannot send or receive request");
	}
	shutdown(sd,2);
#ifdef WIN32
	closesocket(sd);
#else
	close(sd);
	if (rfp) fclose(rfp);
	if (wfp) fclose(wfp);
#endif
	if (r) {
	    return r;
	}
    } else if (sd > 0) {
	logWebObjects(LOGCGI_WARNING,"-- Transaction : cannot fdopen on socket.");
	shutdown(sd,2);
#ifdef WIN32
	closesocket(sd);
#else
	close(sd);
#endif
    }
    return NULL;
}
    
/*
 * execApp() - fork & execute the application
 */	
static int execApp(const char *path, const char *dirPath, char **const argv, WOHTTPRequest *request)
{
#ifndef WIN32
    int execstat;
    int pid;
    int i;
    if (path && dirPath) logWebObjects(LOGCGI_INFO,"-- WOAutostart: Starting app %s from path %s",path,dirPath);

#ifndef WIN32 
    argv[0] = (char *)path;  /* fix #77612 - Mach Autostart on OpenStep 4.2 */
#endif

    switch(pid = fork())
    {
    case 0:
  	/* close the child's stdout(=1) descriptor so that the parent (server) 
	   only gets the CGI output from the CGI's stdout and not WebApp's */
	/* close the child's stderr(=2) too. */
	for (i=0; i<FOPEN_MAX-1; i++)
	    close(i);
        open("/dev/null", O_RDONLY);
        dup2(1, 2);

	if (request->isNSAPI) net_close(request->clientsd); /* this only happens with NSAPI !*/

	if (chdir(dirPath)) {
	    logWebObjects(LOGCGI_WARNING,"-- WOAutostart: Error \"%s\" occured while chdir to %s",strerror(errno),dirPath);
	}
	execstat = execv(path,(char **const)argv);
	if (execstat) 
	    logWebObjects(LOGCGI_WARNING,"-- WOAutostart: Error \"%s\" occured while starting %s from %s",strerror(errno),argv[0],path);
	return -1;
    case -1:
	/* error */
	logWebObjects(LOGCGI_WARNING,"-- WOAutostart: Error \"%s\" occured while forking for %s from path %s",strerror(errno),argv[0],path);
	return 0;
    default:
	return pid;
    }
#else
	SECURITY_ATTRIBUTES sa;
	PROCESS_INFORMATION pInfo;
	STARTUPINFO sInfo;
	int argCount = 0;
	int argLength = 0;
	char *commandLine;
	char *initialDir;
	char drive[_MAX_DRIVE];
	char dir[_MAX_DIR];
	HANDLE out;
	HANDLE stdOut;
	HANDLE error;
	HANDLE stdError;

	/* Grab out stdout handle so we can reset it after creating the process */
	stdOut = GetStdHandle(STD_OUTPUT_HANDLE);
	stdError = GetStdHandle(STD_ERROR_HANDLE);
	out = CreateFile( "NUL", GENERIC_READ | GENERIC_WRITE, 0, &sa, CREATE_ALWAYS,
		FILE_ATTRIBUTE_NORMAL, NULL );
	error = CreateFile( "NUL", GENERIC_READ | GENERIC_WRITE, 0, &sa, CREATE_ALWAYS,
		FILE_ATTRIBUTE_NORMAL, NULL );
	if (! SetStdHandle(STD_OUTPUT_HANDLE, out)) {
	    logWebObjects(LOGCGI_WARNING,"-- WOAutostart: Error setting stdout to NUL");
	}
	if (! SetStdHandle(STD_ERROR_HANDLE, error)) {
	    logWebObjects(LOGCGI_WARNING,"-- WOAutostart: Error setting stderr to NUL");
	}
	
	_splitpath (path, &drive[0], &dir[0], NULL, NULL);
	initialDir = (char*)malloc ((strlen (drive) + strlen (dir) + 1) * sizeof (char*));
	sprintf (initialDir, "%s%s", drive, dir);

	sa.nLength = sizeof (SECURITY_ATTRIBUTES);
	sa.lpSecurityDescriptor = NULL;
	sa.bInheritHandle = FALSE;

	memset (&sInfo, 0, sizeof (STARTUPINFO));
	sInfo.cb = sizeof (STARTUPINFO);
	sInfo.lpReserved = NULL;
	sInfo.lpReserved2 = NULL;
	sInfo.cbReserved2 = 0;
	sInfo.lpDesktop = NULL;
	sInfo.dwFlags = STARTF_USESTDHANDLES;
	sInfo.hStdOutput = out;
	sInfo.hStdError = error;

	while (argv[argCount])
		argLength += strlen (argv[argCount++]);

	commandLine = (char*)malloc ((argLength + argCount) * sizeof (char));
	argCount = 0;

	strcpy (commandLine, "");
	while (argv[argCount])
	{
		strcat (commandLine, argv[argCount++]);
		strcat (commandLine, " ");
	}
	commandLine[argLength+argCount-1] = '\0';

	if (CreateProcess (path, commandLine, &sa, &sa, FALSE, NORMAL_PRIORITY_CLASS, NULL, initialDir, &sInfo, &pInfo)) {
	    FlushFileBuffers( out );
	    SetFilePointer( out, 0, NULL, FILE_BEGIN );
	    FlushFileBuffers( error );
	    SetFilePointer( error, 0, NULL, FILE_BEGIN );
    
/* Reset our stdout handle back to the real stdout */
	    if (! SetStdHandle(STD_OUTPUT_HANDLE, stdOut)) {
		logWebObjects(LOGCGI_WARNING,"-- WOAutostart: Error setting back stdout");
	    }
	    if (! SetStdHandle(STD_ERROR_HANDLE, stdError)) {
		logWebObjects(LOGCGI_WARNING,"-- WOAutostart: Error setting back stderr");
	    }
	    free(commandLine);
 	    free(initialDir);
	    return pInfo.dwProcessId;
	}
        free(commandLine);
        free(initialDir);
	return 0;
#endif 
}


static const char *isDir( const char *dirPath)
{
    struct stat statbuf;
   if (!stat(dirPath,&statbuf)) {
   	if (statbuf.st_mode & S_IFDIR) {
	    return dirPath;
	}
    }
    return NULL;
}


static const char * isExecutable( const char *appPath)
{
    struct stat statbuf;
   if (!stat(appPath,&statbuf)) {
	if ((statbuf.st_mode & S_IFREG) && 
#ifndef WIN32
		/* Windows doesn't set the S_IEXEC bit. */
		(statbuf.st_mode & S_IEXEC) && 
#endif
		!(statbuf.st_mode & S_IFDIR)) {
	    return appPath;
	}
    }
    return NULL;
}

/**
 ** findInRoot: find the app in the root directory
 **
 ** first we find root/WebObjects/appName.woa/appName
 ** then we find root/WebObjects/appName.woa/appName.exe
 ** then we give up
 **
 ** app: name of the app as determined from the URL
 ** root: the directory under which we start our search
 ** dirName: the directory under 'root' where we expect to find the app
 ** appPath (output): the char buffer in which to store the full path
 ** appPathLen: the length of the char buffer appPath
 **/
static char* findInRoot(const char *app, const char *realAppName, const char *root, const char *dirName, char *appPath,int appPathLen)
{
    if (app && *app && root && *root && dirName && *dirName && (appPathLen > 0)) {
	if (((strlen(root) + strlen(dirName) + strlen(app)*2 + 8* sizeof(char)) < appPathLen) && realAppName) {
	    /* if the buffer is big enough, and app is  not WOSERVER or WOEOFSERVER */
            sprintf(appPath,"%s/%s/%s%s/%s",root,dirName, app, WOADAPTOR_APP_EXTENSION, realAppName);
	    logWebObjects(LOGCGI_INFO,"-- WOAutoStart: Looking for %s",appPath);
	    if (isExecutable(appPath)) return appPath;
	}
        if (((strlen(root)+strlen(dirName)+strlen(app)+3*sizeof(char)) < appPathLen) && !realAppName) {
            // if realAppName is NULL we just look for the app (DefaultApp type)...
            sprintf(appPath,"%s/%s/%s",root,dirName, app);
            logWebObjects(LOGCGI_INFO,"-- WOAutoStart : Looking for %s",appPath);
            if (isExecutable(appPath)) return appPath;
        }

	
/* fix for bug #65974 :
 * On NT we need to look for the executable with a .exe too !
 */

#ifdef WIN32
 	if (((strlen(root) + strlen(dirName) + strlen(app)*2 + 12* sizeof(char)) < appPathLen) && realAppName) {
	    /* if the buffer is big enough, and app is  not WOSERVER or WOEOFSERVER */
            sprintf(appPath,"%s/%s/%s%s/%s.exe",root,dirName, app, WOADAPTOR_APP_EXTENSION, realAppName);
	    logWebObjects(LOGCGI_INFO,"-- WOAutoStart: Looking for %s",appPath);
	    if (isExecutable(appPath)) return appPath;
        }
        if (((strlen(root)+strlen(dirName)+strlen(app)+3*sizeof(char)) < appPathLen) && !realAppName) {
            // if realAppName is NULL we just look for the app (DefaultApp type)...
            sprintf(appPath,"%s/%s/%s.exe",root,dirName, app);
            logWebObjects(LOGCGI_INFO,"-- WOAutoStart : Looking for %s",appPath);
            if (isExecutable(appPath)) return appPath;
        }
#endif (WIN32)
    }
     
    return NULL;
}

/* This is for the new WOF3.5 project layout.
 * we look under AppName/AppName.woa for an executable called AppName[.exe]
 */

static char* newFindInRoot(const char *app, const char *realAppName, const char *root, const char *dirName, char *appPath,int appPathLen)
{
    if (app && *app && root && *root && dirName && *dirName && (appPathLen > 0)) {
        if (((strlen(root) + strlen(dirName) + strlen(app)*3 + 9* sizeof(char)) < appPathLen) && realAppName) {
/* if the buffer is big enough, and app is  not WOSERVER or WOEOFSERVER */
            sprintf(appPath,"%s/%s/%s/%s%s/%s",root,dirName, app,realAppName, WOADAPTOR_APP_EXTENSION, realAppName);
            logWebObjects(LOGCGI_INFO,"-- WOAutoStart (WOF3.5): Looking for %s",appPath);
            if (isExecutable(appPath)) return appPath;
        }
        if (((strlen(root)+strlen(dirName)+strlen(app)+3*sizeof(char)) < appPathLen) && !realAppName) {
            // if realAppName is NULL we just look for the app (DefaultApp type)...
            sprintf(appPath,"%s/%s/%s",root,dirName, app);
            logWebObjects(LOGCGI_INFO,"-- WOAutoStart (WOF3.5): Looking for %s",appPath);
            if (isExecutable(appPath)) return appPath;
        }


/* fix for bug #65974 :
 * On NT we need to look for the executable with a .exe too !
 */

#ifdef WIN32
        if (((strlen(root) + strlen(dirName) + strlen(app)*3 + 13* sizeof(char)) < appPathLen) && realAppName) {
            /* if the buffer is big enough, and app is  not WOSERVER or WOEOFSERVER */
            sprintf(appPath,"%s/%s/%s/%s%s/%s.exe",root,dirName, app, realAppName, WOADAPTOR_APP_EXTENSION, realAppName);
            logWebObjects(LOGCGI_INFO,"-- WOAutoStart (WOF3.5): Looking for %s",appPath);
            if (isExecutable(appPath)) return appPath;
        }
        if (((strlen(root)+strlen(dirName)+strlen(app)+3*sizeof(char)) < appPathLen) && !realAppName) {
            // if realAppName is NULL we just look for the app (DefaultApp type)...
            sprintf(appPath,"%s/%s/%s.exe",root,dirName, app);
            logWebObjects(LOGCGI_INFO,"-- WOAutoStart (WOF3.5): Looking for %s",appPath);
            if (isExecutable(appPath)) return appPath;
        }
#endif (WIN32)
    }
     
    return NULL;
}

static int WOAutoStart(const char *appName, const char *docRoot, WOHTTPRequest *request) {
    char appPath[WO_MAX_PATH+1];
    char dirPath[WO_MAX_PATH+1];
    const char *argv[7];
    char defaultApp[WO_MAX_PATH+1];
    char *realAppName;
    int i;
    char *execRoot;

/* Potential security hole :
* We refuse to autostart if any appName or docRoot contains '../'
*/

    if ((appName && strstr(appName,"../")) || (docRoot && strstr(docRoot,".."))) {
        logWebObjects(LOGCGI_INFO,"-- WOAutoStart: Refusing to autoStart for security reason, appName '%s' or docRoot '%s' contains '..'",appName, docRoot);
        return 0;
    }
    
   execRoot = (char *)EXECROOT;
    
/* Fix for bug #63782 :
 * we need to point to the exact application name and not to the whole application path
 */
    realAppName = strrchr(appName,'/'); 	/* last / in the application path */
    if (!realAppName) {
	realAppName = (char *)appName;
    } else {
	realAppName++; 				/* to be just after the last / */
    }
    
    /* prepare for the arguments */
    for (i = 0 ;i < 7; i++) {
	argv[i] = NULL;
    }
    argv[0] = realAppName;
    argv[1] = "-browser";
    argv[2] = "OFF";
    if (docRoot) {
        argv[3] = "-d";
        argv[4] = docRoot;
        argv[5] = appName;
    } else {
        argv[3] = appName;
    }

    logWebObjects(LOGCGI_INFO,"-- WOAutoStart: Trying to AutoStart: appName %s docRoot = %s",appName, docRoot);

/* /NextLibrary/WOApps/HelloWorld.woa/HelloWorld
 */
 
     if (findInRoot(appName,realAppName,execRoot,WEBOBJECTSAPPS,appPath,sizeof(appPath))) {
/* now set dirPath to be the same as appPath up to the directory level */
	strcpy(dirPath,appPath); /* safe as dirPath of same size as appPath */
	dirPath[strrchr(appPath,'/') - appPath]='\0';
	return execApp(appPath, dirPath, (char **const) argv, request);
    }

/* <document-root>/WebObjects/HelloWorld.woa/HelloWorld
*/

     if ((docRoot) && findInRoot(appName,realAppName,docRoot,WEBOBJECTS,appPath,sizeof(appPath))) {
    /* now set dirPath to be the same as appPath up to the directory level */
        strcpy(dirPath,appPath); /* safe as dirPath of same size as appPath */
        dirPath[strrchr(appPath,'/') - appPath]='\0';
        return execApp(appPath, dirPath, (char **const) argv, request);
    }

/* We have to do the next search, to make sure that DefaultApp is there,
 * and to find out if we have a DefaultApp.exe only on NT.
 * /NextLibrary/Executables/DefaultApp(.exe) */	
     if (findInRoot(WOSERVER, NULL, execRoot, EXECUTABLES, appPath,sizeof(appPath))) {
         sprintf(defaultApp,WOSERVER);

#ifdef WIN32
/* fix for bug #65974 :
 * On NT we looked for the executable with a .exe. If that's the one
 * that was found, we change the DefaultApp to DefaultApp.exe !
 */
	if (strstr(appPath,".exe")) {
	    sprintf(defaultApp,".exe");
	}
#endif

	argv[0] = defaultApp;
	
/* FIX FOR BUG #61946 :
 * we now prevent autostarting with DefaultApp if there is no directory of the given appName such as :
 */

 /* <htdocs>/WebObjects/HelloWorld.woa */	
	if ((strlen(docRoot)+strlen(WEBOBJECTS)+strlen(appName)+7)*sizeof(char) < sizeof(dirPath)) {
            sprintf(dirPath, "%s/%s/%s%s",docRoot,WEBOBJECTS,appName, WOADAPTOR_APP_EXTENSION);
	    logWebObjects(LOGCGI_INFO,"-- WOAutoStart: Looking for dir %s",dirPath);
	    if (isDir(dirPath)) {
		return execApp(appPath,dirPath,(char **const) argv, request);
		}
	}

 /* /NextLibrary/WOApps/HelloWorld.woa */
	if ((strlen(execRoot)+strlen(WEBOBJECTSAPPS)+strlen(appName)+3)*sizeof(char) < sizeof(dirPath)) {
            sprintf(dirPath, "%s/%s/%s%s",execRoot,WEBOBJECTSAPPS,appName, WOADAPTOR_APP_EXTENSION);
	    logWebObjects(LOGCGI_INFO,"-- WOAutoStart: Looking for dir %s",dirPath);
	    if (isDir(dirPath)) {
		return execApp(appPath,dirPath,(char **const) argv, request);
		}
	}

	
        logWebObjects(LOGCGI_WARNING,"-- WOAutoStart: Refusing to autostart %s. Could not find any %s%s directory in %s/%s/ or %s/%s/",appName,appName,WOADAPTOR_APP_EXTENSION,execRoot,WEBOBJECTSAPPS,docRoot,WEBOBJECTS);
	
    } else {
        logWebObjects(LOGCGI_WARNING,"-- WOAutoStart: Refusing to autostart %s. Could not find any %s[.exe] in directory %s/%s",appName,WOSERVER,execRoot,EXECUTABLES);
    }

/* New scheme for WOF3.5 
 * The previous code is left for backward compatibility.
 * Now we will look successively for :
 * /NextLibrary/WOApps/HelloWorld/HelloWorld.woa/HelloWorld[.exe]
 * <document-root>/WebObjects/HelloWorld.woa/HelloWorld[.exe]
 * WODefaultApp[.exe] with /NextLibrary/WOApps/HelloWorld/HelloWorld
 * WODefaultApp[.exe] with <document-root>/WebObjects/HelloWorld/HelloWorld
 */

    logWebObjects(LOGCGI_INFO,"-- WOAutoStart: Trying to AutoStart With WOF3.5 project layout : appName %s docRoot = %s",appName, docRoot);

/* /NextLibrary/WOApps/HelloWorld.woa/HelloWorld
 */

    if (newFindInRoot(appName,realAppName,execRoot,WEBOBJECTSAPPS,appPath,sizeof(appPath))) {
/* now set dirPath to be the same as appPath up to the directory level */
        strcpy(dirPath,appPath); /* safe as dirPath of same size as appPath */
        dirPath[strrchr(appPath,'/') - appPath]='\0';
        return execApp(appPath, dirPath, (char **const) argv, request);
    }

/* <document-root>/WebObjects/HelloWorld/HelloWorld.woa/HelloWorld
*/

    if ((docRoot) && newFindInRoot(appName,realAppName,docRoot,WEBOBJECTS,appPath,sizeof(appPath))) {
/* now set dirPath to be the same as appPath up to the directory level */
        strcpy(dirPath,appPath); /* safe as dirPath of same size as appPath */
        dirPath[strrchr(appPath,'/') - appPath]='\0';
        return execApp(appPath, dirPath, (char **const) argv, request);
    }

/* We have to do the next search, to make sure that DefaultApp is there,
 * and to find out if we have a DefaultApp.exe only on NT.
 * /NextLibrary/Executables/DefaultApp(.exe) */	
    if (newFindInRoot(WOSERVER, NULL, execRoot, EXECUTABLES, appPath,sizeof(appPath))) {
        sprintf(defaultApp,WOSERVER);

#ifdef WIN32
/* fix for bug #65974 :
 * On NT we looked for the executable with a .exe. If that's the one
 * that was found, we change the DefaultApp to DefaultApp.exe !
 */
        if (strstr(appPath,".exe")) {
            sprintf(defaultApp,".exe");
        }
#endif

        argv[0] = defaultApp;

/* FIX FOR BUG #61946 :
        * we now prevent autostarting with DefaultApp if there is no directory of the given appName such as :
        */

        /* <htdocs>/WebObjects/HelloWorld/HelloWorld.woa */	
        if ((strlen(docRoot)+strlen(WEBOBJECTS)+strlen(appName)+7)*sizeof(char) < sizeof(dirPath)) {
            sprintf(dirPath, "%s/%s/%s/%s%s",docRoot,WEBOBJECTS,appName, realAppName, WOADAPTOR_APP_EXTENSION);
            logWebObjects(LOGCGI_INFO,"-- WOAutoStart (WOF3.5): Looking for dir %s",dirPath);
            if (isDir(dirPath)) {
                sprintf(dirPath, "%s/%s/%s%s",docRoot,WEBOBJECTS,appName,WOADAPTOR_APP_EXTENSION);
                return execApp(appPath,dirPath,(char **const) argv, request);
                }
        }

        /* /NextLibrary/WOApps/HelloWorld/HelloWorld.woa */
        if ((strlen(execRoot)+strlen(WEBOBJECTSAPPS)+strlen(appName)+3)*sizeof(char) < sizeof(dirPath)) {
            sprintf(dirPath, "%s/%s/%s/%s%s",execRoot,WEBOBJECTSAPPS,appName, realAppName, WOADAPTOR_APP_EXTENSION);
            logWebObjects(LOGCGI_INFO,"-- WOAutoStart (WOF3.5): Looking for dir %s",dirPath);
            if (isDir(dirPath)) {
                sprintf(dirPath, "%s/%s/%s%s",execRoot,WEBOBJECTSAPPS,appName, WOADAPTOR_APP_EXTENSION);
                return execApp(appPath,dirPath,(char **const) argv, request);
                }
        }


        logWebObjects(LOGCGI_WARNING,"-- WOAutoStart: Refusing to autostart %s. Could not find any %s%s directory in %s/%s/%s or %s/%s/",appName,realAppName,WOADAPTOR_APP_EXTENSION,execRoot,WEBOBJECTSAPPS,appName,docRoot,WEBOBJECTS,appName);

    } else {
        logWebObjects(LOGCGI_WARNING,"-- WOAutoStart (WOF3.5): Refusing to autostart %s. Could not find any %s[.exe] in directory %s/%s",appName,WOSERVER,execRoot,EXECUTABLES);
    }

    return 0;
}

int WOIsLoadBalancingEnabled() {
    char *executionRoot = (char *)EXECROOT; // NeXT_ROOT/NextLibrary
    const char *suffix = "Frameworks/WebObjects.framework/Resources/License.table";
    char licenseFile[WO_MAX_PATH+1];
    FILE *fp;
    int isLoadBalancingEnabled = 0;
    char line[1024];
    char key[256];
    char value[256];

    sprintf(licenseFile, "%s/%s", executionRoot, suffix);
    if ( (fp = fopen(licenseFile, "r")) ) {
        while ( getLineFromFilePointer(line, sizeof(line), fp) ) {
            if ( sscanf(line, " %s = %s ", key, value) == 2 ) {
                if ( !strcmp(key, "LoadBalancing") && !strcmp(value, "YES;") )  {
                    isLoadBalancingEnabled = 1;
                }
            }
        }
        if (ferror(fp)) {
            logWebObjects(LOGCGI_WARNING,"-- WOIsLoadBalancingEnabled: Error '%s' occured reading line in license file: %s\n", strerror(errno), licenseFile);
        }
        fclose(fp);
    } else {
        logWebObjects(LOGCGI_WARNING,"-- WOIsLoadBalancingEnabled: Error '%s' occured reading license file: %s\n", strerror(errno), licenseFile);
    }
    return isLoadBalancingEnabled;
}

WOHTTPResponse *TransactionWithAutoStart(const char *appName, int instance, const char *hostName, const char *documentRoot, int timeout, WOHTTPRequest *request, WOURLComponents *urlComponents, const char *httpVersionString, const char *recordingDirPath)
 {
    int pid = 0;
    int port = 0;
    int time = 0;
    char *newHostName = NULL;
    WOHTTPResponse *response;
    int sockDes = -1;
    int status = -1;

    /* As of WOF3.1, we look at a constant location for the Configuration File. This will simplify deployment.
     * EXECROOT is <NEXT_ROOT>/NextLibrary
     * the constant location for the conf file is <NEXT_ROOT>/NextLibrary/WOAdaptors/Configuration/WebObjects.conf
     *
     */

    char configFile[WO_MAX_PATH+1];
    sprintf(configFile, "%s/%s", (char *)EXECROOT, CONFIG_FILE_PATH);
    if (!appName) {
	logWebObjects(LOGCGI_WARNING,"-- TransactionWithAutoStart: appName or hostName is null");
	return NULL;
    }
/* case when no instance is specified, which means either we will load balance, or we will autostart. */

    if (instance == -1) {
        int loadBalancing;

/* We can disable load balancing with the license file. We do this test only with CGI not to impact performance for NSAPI and ISAPI */
        if (!request->isNSAPI && !request->isISAPI && !WOIsLoadBalancingEnabled()) {
            loadBalancing = LOAD_BALANCING_NO_CONFIG;
	} else {
        loadBalancing = WOServerAdaptorGetRandomHostNameAndPort(appName, &instance, &newHostName,&port, &sockDes, timeout, configFile);
	}

	if (loadBalancing == LOAD_BALANCING_FAILURE) {
	    logWebObjects(LOGCGI_WARNING,"-- TransactionWithAutoStart : Load balancing failed");
	    if (newHostName) {
	    	free(newHostName);
	    }
	    return NULL;
	}

/*****************
        if (loadBalancing == LOAD_BALANCING_APPLICATION_NOT_IN_CONFIG) {
            logWebObjects(LOGCGI_WARNING,"-- TransactionWithAutoStart : Load balancing not performed. Autostart not allowed.");
            if (newHostName) {
                free(newHostName);
            }
            return NULL;
        }
****************/
        
/* If we couldn't load balance for one reason or another.
 * we will look in the tmp file to see 
 * if anything was autostarted before under that name, 
 * or we will autostart a new application.
 */
	    
        if ((loadBalancing == LOAD_BALANCING_NO_CONFIG) || (loadBalancing == LOAD_BALANCING_APPLICATION_NOT_IN_CONFIG)) {
	    char hname[256];
	    FILE *lockfp = NULL;
	    int lockfd = -1;
	    int lockautostartfd = -1;

/* we now compare the host name and the server host name...
 * two types of errors are possible :
 */
	    if (gethostname(hname,sizeof(hname))) {
		logWebObjects(LOGCGI_INFO,"-- TransactionWithAutoStart : Error \"%s\" occured while gethostname().",strerror(errno));
		return NULL;
		} 
/* In the second type, hostName is specified and is not localhost and 
 * does not match the actual host name (can't autostart on another host).
 */
#ifdef WIN32
            if (hostName && (stricmp(hname,hostName)) && (stricmp(hostName,"localhost"))){
#else
            if (hostName && (strcasecmp(hname,hostName)) && (strcasecmp(hostName,"localhost"))){
#endif
		logWebObjects(LOGCGI_INFO,"-- TransactionWithAutoStart : Cannot autostart on another host than server host.");
		return NULL;
	    }
	    if (!hostName) {
	    	hostName = hname;
	    }

/* we are on a local host. Let's try to find our tmp config
* Not sure whether we should LockOrFail or LockOrWait here ?
*/
#ifdef WIN32
            lockfd = WOLockOrWait(WOTmpFile());
#else
	    lockfd = WOLockOrWait(TMPFILE);
#endif
	    if (lockfd < 0) {
		logWebObjects(LOGCGI_WARNING,"-- TransactionWithAutoStart : cannot lock config.");
		return NULL;
	    }
#ifdef WIN32
		lockfp = fdopen(_open_osfhandle (lockfd, _O_RDWR | _O_BINARY), "r+");
#else
	    lockfp = fdopen(lockfd,"r+");
#endif
	    if (!lockfp) {
		logWebObjects(LOGCGI_WARNING,"-- TransactionWithAutoStart : cannot fdopen.");
		WOUnlock(lockfd);
		return NULL;
	    }
    
/* instance is always -1 in autostart , so that there is no instance
* now look for the port in the file, and if the port is found look if the process is still on.
* if not autostart.
*/

	    sockDes = WOGetSocketFromPortInTMPConfig(lockfp,appName, hostName, DO_NOT_CHECK_PID, &status);
	    if (status < 0) {
		/* error occured while looking for port in tmp config */
#ifdef WIN32
		logWebObjects(LOGCGI_WARNING,"-- TransactionWithAutoStart : Error occured while looking for port for %s in tmp config %s.",appName,WOTmpFile());
#else
		logWebObjects(LOGCGI_WARNING,"-- TransactionWithAutoStart : Error occured while looking for port for %s in tmp config %s.",appName,TMPFILE);
#endif
		WOUnlock(lockfd);
		fclose(lockfp);
		return NULL;
		}
	    if (status > 0) {
/* Connected : unlock the conf file */
		WOUnlock(lockfd);
		fclose(lockfp);
		}	    
	    else {
/* case when status = 0 (means application not found in tmp config) */
/* We don't want to Autostart if there is no public config file or if the application is neither
 in the public config file or the tmp config file */
                if (loadBalancing == LOAD_BALANCING_APPLICATION_NOT_IN_CONFIG) {
                    logWebObjects(LOGCGI_WARNING,"-- TransactionWithAutoStart : Application not in private nor public config file.  Not autostarting because there is a public config file.");
                    WOUnlock(lockfd);
                    fclose(lockfp);
                    return NULL;
                }
                
/* can't find port in tmp config and host name. Autostart */

/* First, we will need the document root, so there is no point continuing
 * if we don't have it now in the request structure.
 */
 
 if (!dict_findval(DOCUMENT_ROOT, request->headers)) {
	logWebObjects(LOGCGI_WARNING,"-- TransactionWithAutoStart : cannot autostart when the server does not provide the document root. Start application %s by hand, specifying the document root with the -d option on the command line", appName);
        WOUnlock(lockfd);
        fclose(lockfp);
	return NULL;
    }

/* Second lets TestLock the /tmp/WebObjects.autostart file if it exists
 * or create it. This fails if the file is locked or any error occurs.
 */ 
#ifdef WIN32
            lockautostartfd = WOLockOrFail(WOTmpAutoStartFile());
#else
	    lockautostartfd = WOLockOrFail(TMPAUTOSTARTFILE);
#endif
	    if (lockautostartfd < 0) {
		logWebObjects(LOGCGI_WARNING,"-- TransactionWithAutoStart : cannot lock autostart file %s", TMPAUTOSTARTFILE);
/* Do not forget to unlock conf file if autostart fails*/
		WOUnlock(lockfd);
		fclose(lockfp);
		return WOErrorResponseWithFile(documentRoot, ERROR_RESPONSE_LOCKED_AUTOSTART_FILE);
	    }

		logWebObjects(LOGCGI_INFO,"-- TransactionWithAutoStart : Autostarting app:%s on host %s",appName,hostName);

/* now we unlock the tmpfile, so that the application can write in it, but we keep the autostart tmp file
* locked, so that no one else can autostart in the mean time...
*/

		WOUnlock(lockfd);
		fclose(lockfp);

/* now we do the autostart */

		pid = WOAutoStart(appName, documentRoot, request);
		if (pid >0) {
		    logWebObjects(LOGCGI_INFO,"-- TransactionWithAutoStart : started process %d. Synchronizing...",pid);
		    
/* now synchronization loop until we find a match in the tmpfile (matching pid) */
		    do {
		    
/* We noticed that in some instances sleep crashes on solaris... */
#if (NeXT_PDO && __svr4__)
			sleep2((unsigned int)1);
#else
			sleep((unsigned int)1); 
#endif
			time ++;
			if (time > timeout) {
/* At timeout don't forget to unlock the autostart file */
			    logWebObjects(LOGCGI_WARNING,"-- TransactionWithAutoStart : exceeded timeout while waiting for autostarted application to come up\n");
			    WOUnlock(lockautostartfd);
                            return WOErrorResponseWithFile(documentRoot, ERROR_RESPONSE_WAITED_TOO_LONG);
			}
			    
/* relock the tmp file */
#ifdef WIN32
			lockfd = WOLockOrWait(WOTmpFile());
#else
			lockfd = WOLockOrWait(TMPFILE);
#endif
			if (lockfd < 0) {
/* if locking error unlock the autostart file and bail out */
			    logWebObjects(LOGCGI_WARNING,"-- TransactionWithAutoStart : cannot lock config.");
			    WOUnlock(lockautostartfd);
                            return WOErrorResponseWithFile(documentRoot, ERROR_RESPONSE_CANNOT_LOCK_CONFIG);
			}

/* reopen the tmp file */
#ifdef WIN32
			lockfp = fdopen(_open_osfhandle (lockfd, _O_RDWR | _O_BINARY), "r+");
#else
			lockfp = fdopen(lockfd,"r+");
#endif
			if (!lockfp) {
/* if opening error unlock the autostart file and bail out */
			    logWebObjects(LOGCGI_WARNING,"-- TransactionWithAutoStart : cannot fdopen.");
			    WOUnlock(lockfd);				
			    fclose(lockfp);
			    WOUnlock(lockautostartfd);
                            return WOErrorResponseWithFile(documentRoot, ERROR_RESPONSE_CANNOT_OPEN_CONFIG);
			}

/* now look for the new pid, which means the app is on and running. */
			sockDes = WOGetSocketFromPortInTMPConfig(lockfp, appName, hostName, pid, &status);
			if (status < 0) {
/* error occured while looking for port in tmp config
 * don't forget to unlock the autostart file and the
 * conf file before bailing out 
 */
#ifdef WIN32
			    logWebObjects(LOGCGI_WARNING,"-- TransactionWithAutoStart : Error occured while looking for port for %s in tmp config %s.",appName, WOTmpFile());
#else
			    logWebObjects(LOGCGI_WARNING,"-- TransactionWithAutoStart : Error occured while looking for port for %s in tmp config %s.",appName,TMPFILE);
#endif
			    WOUnlock(lockfd);				
			    fclose(lockfp);
			    WOUnlock(lockautostartfd);
                            return WOErrorResponseWithFile(documentRoot, ERROR_RESPONSE_PORT_NOT_IN_CONFIG);
			    }
/* now we unlock the tmpfile, so that the application can write in it, 
 * but we keep the autostart tmp file
 * locked, so that no one else can autostart in the mean time...
 */
			WOUnlock(lockfd);				
			fclose(lockfp);

		    }
/* as long as we didn't find the port */
		    while (!status); 
		}
	    else {
/* Couldn't find the application or starting the app failed */
	    	logWebObjects(LOGCGI_WARNING,"-- TransactionWithAutoStart : Bailing out of autostart.");
		WOUnlock(lockautostartfd);
#ifdef WODIRECTORYLISTING
                return WODirectoryListing(appName,documentRoot,urlComponents);
#else
		return NULL;
#endif
		}		

/* we found the port, we can unlock the autostart file, so that other clients can try to autostart.
    */
	WOUnlock(lockautostartfd);
	}
    }
    else {
	if (hostName) {
		free((char *)hostName);
		hostName=NULL;
	}
    	hostName = newHostName;
	if (sockDes < 0) {
		logWebObjects(LOGCGI_WARNING,"-- WOTransaction : Bad socket descriptor after load balancing.");
	    return NULL;
	}
    }
	
            response = Transaction(appName,instance,sockDes,hostName,timeout,request,urlComponents,httpVersionString, recordingDirPath);
    if (newHostName) {
	    free(newHostName);
	    newHostName = NULL;
	    }
    return response;
    }
    else {
        //Here is the case when the url came in with an instance number.
    	char hname[256];

        if (!request->isNSAPI && !request->isISAPI && !WOIsLoadBalancingEnabled()) {
            logWebObjects(LOGCGI_WARNING,"-- WOTransaction : Illegal url. Request handling by a remote host is disabled in this version of WebObjects.");
            return NULL;
        }

	if (!hostName) {
	    gethostname(hname,sizeof(hname));
	    hostName = hname;
	}		
	
	port = WOGetPortFromConfigFile(appName,instance,(const char *)hostName,configFile);
	if (port <= 0) {
	    return NULL;
	    }
	logWebObjects(LOGCGI_INFO,"-- WOTransaction : connecting to app %s on host %s with port %d", appName, hostName, port);
	sockDes = WOSocketConnectTo(hostName, port, timeout);  
	if (sockDes < 0) {
		logWebObjects(LOGCGI_WARNING,"-- WOTransaction : Connection failed on port %d for %s:%d on host %s", port, appName, instance, hostName);
	    return NULL;
	}
        return Transaction(appName,instance,sockDes,hostName,timeout,request,urlComponents,httpVersionString, recordingDirPath);
    }
    
return NULL;
}
