/*
    WOURLCUtilities.h
    Copyright (c) 1995-1996, NeXT Software, Inc.  All rights reserved.

    The utilities defined in this file help with the manipulation of WOF dynamic URLs.
    WOF dynamic URLs are composed of the following components:
	<APPLICATION-NAME>: The name of the application the URL refers to.
	<PAGE-NAME>: The name of the page in which the URL was generated.
	<SESSION-ID>: The ID of the session in which the URL was generated.
	<CONTEXT-ID>: The ID of the context in which the URL was generated.
	<SENDER-ID>: The ID of the element associated with the URL.
	<APPLICATION-NUMBER>: The instance number of the URL's application.
	<APPLICATION-HOST>: The name of the host running the URL's application.
	<WEBOBJECTS-VERSION>: A string identifying the version of WebObjects run by the URL's application.
	<PREFIX>: A server-specific string.
    These utilitie allow parsing of a WOF dynamic URL string for its components and generation of a WOF dynamic URL string from its components.  Because the format of a WOF dynamic URL may change over time, these utilities should always be used to access a WOF dynamic URL's components or to generate a WOF dynamic URL (i.e. no assumptions should be made about a WOF dynamic URL's format).
*/

#ifndef _WOF_URL_C_UTILITIES_H
#define _WOF_URL_C_UTILITIES_H

/*********** WOF dynamic URL types. ***********/

typedef struct _WOURLComponent {
    const char *start;
    unsigned int length;
} WOURLComponent;

typedef struct _WOURLComponents {
    WOURLComponent prefix;
    WOURLComponent webObjectsVersion;
    WOURLComponent applicationName;
    WOURLComponent applicationNumber;
    WOURLComponent applicationHost;
    WOURLComponent sessionID;
    WOURLComponent pageName;
    WOURLComponent contextID;
    WOURLComponent senderID;
    WOURLComponent queryString;
    WOURLComponent suffix;
} WOURLComponents;

typedef enum {
    WOURLOK = 0,
    WOURLInvalidPrefix = 1,
    WOURLInvalidWebObjectsVersion = 2,
    WOURLInvalidApplicationName = 3,
    WOURLInvalidApplicationNumber = 4,
    WOURLInvalidApplicationHost = 5,
    WOURLInvalidPageName = 6,
    WOURLInvalidSessionID = 7,
    WOURLInvalidContextID = 8,
    WOURLInvalidSenderID = 9,
    WOURLInvalidQueryString = 10,
    WOURLInvalidSuffix = 11
} WOURLError;

/*********** WOF dynamic URL functions. ***********/

void WOParseURL(WOURLComponents *components, const char *string);
    /* ... */

WOURLError WOCheckURL(WOURLComponents *components);
    /* ... */

unsigned int WOSizeURL(WOURLComponents *components);
    /* ... */

void WOComposeURL(char *string, WOURLComponents *components);
    /* ... */

#endif /* _WOF_URL_C_UTILITIES_H */
