/* ====================================================================
 * Copyright (c) 1995-1997 The Apache Group.  All rights reserved.
 *
 * Redistribution and use in source and binary forms, with or without
 * modification, are permitted provided that the following conditions
 * are met:
 *
 * 1. Redistributions of source code must retain the above copyright
 *    notice, this list of conditions and the following disclaimer. 
 *
 * 2. Redistributions in binary form must reproduce the above copyright
 *    notice, this list of conditions and the following disclaimer in
 *    the documentation and/or other materials provided with the
 *    distribution.
 *
 * 3. All advertising materials mentioning features or use of this
 *    software must display the following acknowledgment:
 *    "This product includes software developed by the Apache Group
 *    for use in the Apache HTTP server project (http://www.apache.org/)."
 *
 * 4. The names "Apache Server" and "Apache Group" must not be used to
 *    endorse or promote products derived from this software without
 *    prior written permission.
 *
 * 5. Redistributions of any form whatsoever must retain the following
 *    acknowledgment:
 *    "This product includes software developed by the Apache Group
 *    for use in the Apache HTTP server project (http://www.apache.org/)."
 *
 * THIS SOFTWARE IS PROVIDED BY THE APACHE GROUP ``AS IS'' AND ANY
 * EXPRESSED OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE
 * IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR
 * PURPOSE ARE DISCLAIMED.  IN NO EVENT SHALL THE APACHE GROUP OR
 * ITS CONTRIBUTORS BE LIABLE FOR ANY DIRECT, INDIRECT, INCIDENTAL,
 * SPECIAL, EXEMPLARY, OR CONSEQUENTIAL DAMAGES (INCLUDING, BUT
 * NOT LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES;
 * LOSS OF USE, DATA, OR PROFITS; OR BUSINESS INTERRUPTION)
 * HOWEVER CAUSED AND ON ANY THEORY OF LIABILITY, WHETHER IN CONTRACT,
 * STRICT LIABILITY, OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE)
 * ARISING IN ANY WAY OUT OF THE USE OF THIS SOFTWARE, EVEN IF ADVISED
 * OF THE POSSIBILITY OF SUCH DAMAGE.
 * ====================================================================
 *
 * This software consists of voluntary contributions made by many
 * individuals on behalf of the Apache Group and was originally based
 * on public domain software written at the National Center for
 * Supercomputing Applications, University of Illinois, Urbana-Champaign.
 * For more information on the Apache Group and the Apache HTTP server
 * project, please see <http://www.apache.org/>.
 *
 */

/*
 * Apache WebObjects Adaptor, (c)1997-1998 by Gabriel Sebestyen
 */

#include "httpd.h"
#include "http_config.h"
#include "http_core.h"
#include "http_log.h"
#include "http_main.h"
#include "http_protocol.h"
#include "util_script.h"



#include <stdio.h>
#include <stdlib.h>

#include <unistd.h>

#include <string.h>
#include <sys/param.h> 

#include "WODict.h"
#include "WOHTTPProtocol.h"
#include "WOAdaptorUtil.h"
#include "WOTransaction.h"
#include "WOURLCUtilities.h"
#include "WODirectoryListing.h"
#include "WORecording.h"


#define TIME_TEST 0 
#define LOAD_BALANCING 1 

#define QUOTE		'"'
	


extern char *WO_CGI_Headers[32];
extern char *WO_Req_Headers[32];



#define	TGET(a)	ap_table_get(r->subprocess_env, a )


module wo_module;

/* KLUDGE --- for back-combatibility, we don't have to check ExecCGI
 * in ScriptAliased directories, which means we need to know if this
 * request came through ScriptAlias or not... so the Alias module
 * leaves a note for us.
 */


#define RDIE(a)								\
    r->content_type = ap_pstrdup (r->pool, "text/html");			\
    ap_rprintf(r,"<html><body><strong>%s</strong></body></html>\n", (char *)a );	\
    ap_rflush(r);								\
    return OK;



/* 
 * get_documentRoot() - return document root. 
 */
static char *get_documentRoot(char *pathTranslated, char *pathInfo,char *buffer, int buflen)
{    
    if (pathTranslated && pathInfo) {
	int len;


	len = strlen(pathTranslated) - strlen(pathInfo);
	if(len >0 && (!strcmp(&pathTranslated[len],pathInfo))) { 
	    /* There is a possible security thing. If my doc root is
		/me/PrivateSecurePlace/importantThings/www, and then
		the CGI gives me
		/me/PrivateSecurePlace/importThings/www/foo/bar as 
		pathTranslated, and then /importThings/www/foo/bar as
		pathInfo, then they might be able to access things in
		/me/PrivateSecurePlace. But in real life (tm) that shouldnt
		really happen. If they hacked the server you got other
		problems */
	    if (len < buflen) {
		strncpy(buffer,pathTranslated,len);
		buffer[len] = '\0';
		return buffer;
	    }
	}
    } 
    buffer[0]='\0';

    return buffer;
}    


int wo_handler (request_rec *r)
{
    int retval;
    char *argv0;

    char *path_info, *server_protocol, *script_name, *p;
    char *urlString;
    WOHTTPRequest *request = NULL;
    WOHTTPResponse *response;
    int appInstance=-1;
    WOURLComponents *url = NULL;
    char recordingDirPath[WO_MAX_PATH+1];

/* intermediate stuff */
    char *appName = NULL;
    char *hostName = NULL;
    char documentRoot[MAXPATHLEN+1];
    int method;
    int len;
    char **cgi_header, **req_header;

/* get content stuff */
    void *qs, *contentLength;
    int inputLen;


    logWebObjects(LOGCGI_INFO,"-- AWOA -- Starting --");

    if (r->method_number == M_OPTIONS) {
        /* 99 out of 100 CGI scripts, this is all they support */
        r->allowed |= (1 << M_GET);
        r->allowed |= (1 << M_POST);
	return DECLINED;
    }


    if((argv0 = strrchr(r->filename,'/')) != NULL)
        argv0++;
    else argv0 = r->filename;

    //Is nph needed?
    //nph = !(strncmp(argv0,"nph-",4));


    if ((retval = ap_setup_client_block(r, REQUEST_CHUNKED_ERROR)))
	return retval;
    //Needed to change
    ap_add_common_vars(r);
    ap_add_cgi_vars(r);


//Ide jon a wo blokk
    if ( (r->method_number != M_GET) && (r->method_number != M_PUT) && (r->method_number != M_POST) ) {
	RDIE("Unknown HTTP method, or HTTP PUT or HEAD method not supported.")
    }

    path_info = r->path_info;
    if (!path_info) {
	RDIE("No application name and other arguments in URL, or path_info too long")
    }
    
    server_protocol = r->protocol;
    if (!server_protocol) {
	RDIE("Cannot determine server protocol from CGI")
    }

    script_name = TGET(CGI_SCRIPT_NAME);	//"SCRIPT_NAME"
    if (!script_name) {
	RDIE("Cannot determine script name from CGI")
    }


    logWebObjects(LOGCGI_INFO,"-- AWOA: script_name = %s\n\tserver_protocol = %s\n\tpath_info = %s",script_name, server_protocol, path_info);

    /* cheating ... */
    len = strlen(script_name)+strlen(path_info);
    urlString = malloc((len+1)*sizeof(char));
    sprintf(urlString,"%s%s",script_name,path_info);

/* Do not worry, this will only do something in the case of the WebObjects-Recording adaptor */
    if (getRecordingDir(TGET(CGI_QUERY_STRING), recordingDirPath, sizeof(recordingDirPath)) < 0) {
	response = WOErrorResponseWithFile(documentRoot, ERROR_RESPONSE_NO_RECORDING_FILE);
    }

    url = (WOURLComponents *)malloc(sizeof(WOURLComponents));
    WOParseURL(url, urlString);
    switch(WOCheckURL(url)) {
	case WOURLOK:
	    break;
	case WOURLInvalidPrefix:
	    RDIE("Invalid prefix in URL")
	    break;
	case WOURLInvalidApplicationName:
	    RDIE("Invalid application name in URL")
	    break;
	case WOURLInvalidApplicationNumber:
	    RDIE("Invalid application number in URL")
	    break;
	case WOURLInvalidApplicationHost:
	    RDIE("Invalid application host name in URL")
	    break;
	case WOURLInvalidPageName:
	    RDIE("Invalid page name in URL")
	    break;
	case WOURLInvalidSessionID:
	    RDIE("Invalid session ID in URL")
	    break;
	case WOURLInvalidContextID:
	    RDIE("Invalid context ID in URL")
	    break;
	case WOURLInvalidSenderID:
	    RDIE("Invalid sender ID in URL")
	    break;
	case WOURLInvalidQueryString:
	    RDIE("Invalid query string in URL")
	    break;
	default:
	    RDIE("Invalid URL")
	    break;
    }


    if (url->applicationName.length) {
	    appName = malloc((url->applicationName.length+1)*sizeof(char));
	    strncpy(appName,url->applicationName.start,url->applicationName.length);
	    appName[url->applicationName.length] = '\0';
	    }
    if (url->applicationHost.length) {
	    hostName = malloc((url->applicationHost.length+1)*sizeof(char));
	    strncpy(hostName,url->applicationHost.start,url->applicationHost.length);
	    hostName[url->applicationHost.length] = '\0';
	    }
    if (url->applicationNumber.length) {
	    char *appInstanceString;
	    appInstanceString = malloc((url->applicationNumber.length+1)*sizeof(char));
	    strncpy(appInstanceString,url->applicationNumber.start,url->applicationNumber.length);
	    appInstanceString[url->applicationNumber.length] = '\0';
	    appInstance = atol(appInstanceString);
	    free(appInstanceString);
    }


    if (!appName) {
	RDIE("Cannot obtain the name of the application being requested, or the application name is too long")
    }

    get_documentRoot(TGET(CGI_PATH_TRANSLATED), path_info, documentRoot, sizeof(documentRoot));
    
    logWebObjects(LOGCGI_INFO,"-- AWOA: \tAppName = %s\n\tInstance = %d\n\tHostName = %s\n\tdocumentRoot = %s",appName ? appName : "NONE", appInstance, hostName ? hostName : "NONE", (documentRoot && *documentRoot) ? documentRoot : "NONE");

/* 
 * Build the request
 */
    request = (WOHTTPRequest *) malloc(sizeof(WOHTTPRequest));


    request->headers = dict_create(30);

    /* Adds the private document root header. It will be removed at the adaptor level in the
     * application adaptor.
     */

    if (documentRoot && *documentRoot) {
	dict_nvinsert(DOCUMENT_ROOT,documentRoot,request->headers);
    }

    /* fill in all the HTTP related information */
    cgi_header = WO_CGI_Headers;
    req_header = WO_Req_Headers;

    while (*cgi_header && *req_header) {
            if ((p = TGET(*cgi_header))) {
                dict_nvinsert(*req_header,p,request->headers);
            }
        cgi_header++;
        req_header++;
    }

/* Build the request line */

    if (r->method_number == M_GET) {
        char *qs = TGET(CGI_QUERY_STRING);
        url->queryString.start = qs;
        url->queryString.length = qs ? strlen(qs) : 0;
        request->content = NULL;
    } else if (r->method_number == M_POST) {
/* get content */
	contentLength = TGET(CGI_CONTENT_LENGTH);
	if (!contentLength) request->content = NULL;
	else {
	    inputLen = atol(contentLength);
	}
	qs =(void *)malloc(inputLen*sizeof(char)); 	

	if ( ap_bread( r->connection->client, qs, inputLen ) != inputLen) {
	    /* I have to punt here. THE CGI doc says you have to have the right size or you suck. And if I dont get the right size, I am not sure the info I am getting is parsable. */
	    free(qs);
	    request->content = NULL;
	} else {
	    request->content = qs;
	}
    } else {
	request->content = NULL;
    }

    request->isNSAPI = 0; /* just to say that we are not using netscape api here */
    request->isISAPI = 0; /* just to say that we are not using ISAPI here */
    response = TransactionWithAutoStart(appName,appInstance,hostName, documentRoot, 300, request, url, server_protocol, recordingDirPath);
    if (!response) {
        RDIE("Did not receive any response from application. It is possible that the application does not exist, or that the requested url is incorrect.")
    }

{
	char *version, *msg, *newline, *length;
	int stat,i;
    
	len = strlen(response->statusLine);
	version = malloc((len+1)*sizeof(char));
	msg = malloc((len+1)*sizeof(char));
	newline = malloc((len*2+1)*sizeof(char));
	sscanf(response->statusLine,"%s %d %s",version,&stat,msg);

	r->status_line = ap_palloc(r->pool, len);
	sprintf(r->status_line, "%d %s", stat, msg);
	r->status = stat;

	for (i=0;i < response->headers->hsize;i++) { /* headers */
	    struct dict_entry *dictEntry;

	    dictEntry = response->headers->ht[i];
	    while (dictEntry) {
		dict_param *dictParam;

		dictParam = dictEntry->param;
		if (dictParam && dictParam->name && dictParam->value) {

		    if (!strcasecmp(dictParam->name ,"Content-Type")) {
			r->content_type = ap_pstrdup (r->pool, dictParam->value);
		    } else if (!strcasecmp(dictParam->name, "Status")) {
			sscanf(dictParam->value, "%d", &r->status);
			r->status_line = ap_pstrdup(r->pool, dictParam->value);
		    } else if (!strcasecmp(dictParam->name, "Location")) {
			ap_table_set (r->headers_out, "Location", dictParam->value);
		    }  else if (!strcasecmp(dictParam->name, "Content-Length")) {
			ap_table_set (r->headers_out, "Content-Length", dictParam->value);
		    } else if (!strcasecmp(dictParam->name, "Transfer-Encoding")) {
			ap_table_set (r->headers_out, "Transfer-Encoding", dictParam->value);
		    } else if (!strcasecmp(dictParam->name, "Set-Cookie")) {
			ap_table_add(r->err_headers_out, "Set-Cookie", dictParam->value);
		    } else {
			ap_table_merge(r->err_headers_out, dictParam->name, dictParam->value);
		    }
		}
		dictEntry = dictEntry->next;
		if (!dictEntry) break;
	    }
	}
	
	ap_send_http_header(r);

	if ((length = dict_findval(CONTENT_LENGTH,response->headers))) {
	    int contentLength = atoi(length);
	    if (contentLength) {	/* Write the content */
		ap_rwrite(response->content, contentLength, r);
	    }
	}
        if (version) {
            free(version); 
        }
        if (msg) {
            free(msg);
        }
        if (newline) {
            free(newline);
        }                     
    }

    ap_rflush(r);  /* flush it */

/* no longer needed */
    free(urlString);
    free(response);
    if (appName) {
        free(appName);
    }
    if (hostName) {
        free(hostName);
    }
    if (url) {
        free(url);
    }
    if (request) {
        free(request);
    }
//Eddig

    logWebObjects(LOGCGI_INFO,"-- AWOA -- Ended --");
    return OK;
}



handler_rec wo_handlers[] = {
{ "wo-adaptor", wo_handler },
{ "wo-exec", wo_handler },
{ NULL }
};

module wo_module = {
   STANDARD_MODULE_STUFF,
   NULL,			/* initializer */
   NULL,			/* dir config creater */
   NULL,			/* dir merger --- default is to override */
   NULL,			/* server config */
   NULL,		       	/* merge server config */
   NULL,			/* command table */
   wo_handlers,			/* handlers */
   NULL,			/* filename translation */ 
   NULL,			/* check_user_id */
   NULL,			/* check auth */
   NULL,			/* check access */
   NULL,			/* type_checker */
   NULL,			/* fixups */
   NULL,			/* logger */
   NULL				/* header parser */
};
